#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date, datetime

from dateutil import tz
from dateutil.relativedelta import MO
from dateutil.relativedelta import relativedelta as rd
from pymeeus.Epoch import Epoch
from pymeeus.Sun import Sun

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import OCT, NOV, DEC
from holidays.holiday_base import HolidayBase


class Japan(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Japan
    """

    country = "JP"
    special_holidays = {
        1959: ((APR, 10, "結婚の儀"),),  # The Crown Prince marriage ceremony.
        1989: ((FEB, 24, "大喪の礼"),),  # State Funeral of Emperor Shōwa.
        1990: ((NOV, 12, "即位礼正殿の儀"),),  # Enthronement ceremony.
        1993: ((JUN, 9, "結婚の儀"),),  # The Crown Prince marriage ceremony.
        2019: (
            (MAY, 1, "天皇の即位の日"),  # Enthronement day.
            (OCT, 22, "即位礼正殿の儀が行われる日"),  # Enthronement ceremony.
        ),
    }

    def _populate(self, year):
        if year < 1949 or year > 2099:
            raise NotImplementedError

        super()._populate(year)

        # New Year's Day
        self[date(year, JAN, 1)] = "元日"

        # Coming of Age Day
        if year <= 1999:
            self[date(year, JAN, 15)] = "成人の日"
        else:
            self[date(year, JAN, 1) + rd(weekday=MO(+2))] = "成人の日"

        # Foundation Day
        if year >= 1967:
            self[date(year, FEB, 11)] = "建国記念の日"

        # Reiwa Emperor's Birthday
        if year >= 2020:
            self[date(year, FEB, 23)] = "天皇誕生日"

        # Vernal Equinox Day
        epoch = Sun.get_equinox_solstice(year, target="spring")
        equinox = map(int, Epoch(epoch).get_full_date())
        adjusted_date = datetime(*equinox, tzinfo=tz.UTC).astimezone(
            tz.gettz("Asia/Tokyo")
        )
        self[date(year, MAR, adjusted_date.day)] = "春分の日"

        # Showa Emperor's Birthday, Greenery Day or Showa Day
        if year <= 1988:
            self[date(year, APR, 29)] = "天皇誕生日"
        elif year <= 2006:
            self[date(year, APR, 29)] = "みどりの日"
        else:
            self[date(year, APR, 29)] = "昭和の日"

        # Constitution Memorial Day
        self[date(year, MAY, 3)] = "憲法記念日"

        # Greenery Day
        if year >= 2007:
            self[date(year, MAY, 4)] = "みどりの日"

        # Children's Day
        self[date(year, MAY, 5)] = "こどもの日"

        # Marine Day
        if 1996 <= year <= 2002:
            self[date(year, JUL, 20)] = "海の日"
        elif year == 2020:
            self[date(year, JUL, 23)] = "海の日"
        elif year == 2021:
            self[date(year, JUL, 22)] = "海の日"
        elif year >= 2003:
            self[date(year, JUL, 1) + rd(weekday=MO(+3))] = "海の日"

        # Mountain Day
        if year == 2020:
            self[date(year, AUG, 10)] = "山の日"
        elif year == 2021:
            self[date(year, AUG, 8)] = "山の日"
        elif year >= 2016:
            self[date(year, AUG, 11)] = "山の日"

        # Respect for the Aged Day
        if 1966 <= year <= 2002:
            self[date(year, SEP, 15)] = "敬老の日"
        elif year >= 2003:
            self[date(year, SEP, 1) + rd(weekday=MO(+3))] = "敬老の日"

        # Autumnal Equinox Day
        epoch = Sun.get_equinox_solstice(year, target="autumn")
        equinox = map(int, Epoch(epoch).get_full_date())
        adjusted_date = datetime(*equinox, tzinfo=tz.UTC).astimezone(
            tz.gettz("Asia/Tokyo")
        )
        self[date(year, SEP, adjusted_date.day)] = "秋分の日"

        # Health and Sports Day
        if 1966 <= year <= 1999:
            self[date(year, OCT, 10)] = "体育の日"
        elif 2000 <= year <= 2019:
            self[date(year, OCT, 1) + rd(weekday=MO(+2))] = "体育の日"
        elif year == 2020:
            self[date(year, JUL, 24)] = "スポーツの日"
        elif year == 2021:
            self[date(year, JUL, 23)] = "スポーツの日"
        elif 2022 <= year:
            self[date(year, OCT, 1) + rd(weekday=MO(+2))] = "スポーツの日"

        # Culture Day
        self[date(year, NOV, 3)] = "文化の日"

        # Labour Thanksgiving Day
        self[date(year, NOV, 23)] = "勤労感謝の日"

        # Regarding the Emperor of Heisei
        if 1989 <= year <= 2018:
            # Heisei Emperor's Birthday
            self[date(year, DEC, 23)] = "天皇誕生日"

        # A weekday between national holidays becomes a holiday too (国民の休日)
        self._add_national_holidays(year)

        # Substitute holidays
        self._add_substitute_holidays(year)

    def _add_national_holidays(self, year):
        if year in {
            1988,
            1989,
            1990,
            1991,
            1993,
            1994,
            1995,
            1996,
            1999,
            2000,
            2001,
            2002,
            2004,
            2005,
            2006,
        }:
            self[date(year, MAY, 4)] = "国民の休日"

        if year in {2032, 2049, 2060, 2077, 2088, 2094}:
            self[date(year, SEP, 21)] = "国民の休日"
        elif year in {2009, 2015, 2026, 2037, 2043, 2054, 2065, 2071, 2099}:
            self[date(year, SEP, 22)] = "国民の休日"
        elif year == 2019:
            self[date(year, APR, 30)] = "国民の休日"
            self[date(year, MAY, 2)] = "国民の休日"

    def _add_substitute_holidays(self, year):
        table = (
            (
                1,
                2,
                {
                    1978,
                    1984,
                    1989,
                    1995,
                    2006,
                    2012,
                    2017,
                    2023,
                    2034,
                    2040,
                    2045,
                },
            ),
            (1, 16, {1978, 1984, 1989, 1995}),
            (
                2,
                12,
                {1979, 1990, 1996, 2001, 2007, 2018, 2024, 2029, 2035, 2046},
            ),
            (2, 24, {2020, 2025, 2031, 2042, 2048}),
            (3, 21, {1988, 2005, 2016, 2033, 2044, 2050}),
            (3, 22, {1982, 1999, 2010, 2027}),
            (
                4,
                30,
                {
                    1973,
                    1979,
                    1984,
                    1990,
                    2001,
                    2007,
                    2012,
                    2018,
                    2029,
                    2035,
                    2040,
                    2046,
                },
            ),
            (5, 4, {1981, 1987, 1992, 1998}),
            (
                5,
                6,
                {
                    1974,
                    1985,
                    1991,
                    1996,
                    2002,
                    2013,
                    2019,
                    2024,
                    2030,
                    2041,
                    2047,
                    2008,
                    2014,
                    2025,
                    2031,
                    2036,
                    2042,
                    2009,
                    2015,
                    2020,
                    2026,
                    2037,
                    2043,
                    2048,
                },
            ),
            (7, 21, {1997}),
            (8, 9, {2021}),
            (8, 12, {2019, 2024, 2030, 2041, 2047}),
            (9, 16, {1974, 1985, 1991, 1996, 2002}),
            (9, 23, {2024}),
            (9, 24, {1973, 1984, 1990, 2001, 2007, 2018, 2029, 2035, 2046}),
            (10, 11, {1976, 1982, 1993, 1999}),
            (
                11,
                4,
                {
                    1974,
                    1985,
                    1991,
                    1996,
                    2002,
                    2013,
                    2019,
                    2024,
                    2030,
                    2041,
                    2047,
                },
            ),
            (
                11,
                24,
                {
                    1975,
                    1980,
                    1986,
                    1997,
                    2003,
                    2008,
                    2014,
                    2025,
                    2031,
                    2036,
                    2042,
                },
            ),
            (12, 24, {1990, 2001, 2007, 2012, 2018}),
        )
        for holiday in table:
            month = holiday[0]
            day = holiday[1]
            years = holiday[2]
            if year in years:
                self[date(year, month, day)] = "振替休日"


class JP(Japan):
    pass


class JPN(Japan):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, APR, MAY, JUL, AUG, OCT, NOV, DEC, SUN
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class Burundi(HolidayBase):
    """
    Burundian holidays
    Note that holidays falling on a sunday maybe observed
    on the following Monday.
    This depends on formal annoucemnts by the government,
    which only happens close to the date of the holiday.

    Primary sources:
    https://www.officeholidays.com/countries/burundi
    """

    country = "BI"

    def _populate(self, year):
        super()._populate(year)

        def _add_holiday(dt: date, hol: str) -> None:
            """Only add if in current year; prevents adding holidays across
            years (handles multi-day Islamic holidays that straddle Gregorian
            years).
            """
            if dt.year == year:
                self[dt] = hol

        # New Year's Day
        self[date(year, JAN, 1)] = "New Year's Day"

        # Unity Day
        name = "Unity Day"
        self[date(year, FEB, 5)] = name
        if date(year, FEB, 5).weekday() == SUN:
            self[date(year, FEB, 6)] = name + " (Observed)"

        # President Ntaryamira Day
        name = "President Ntaryamira Day"
        self[date(year, APR, 6)] = "President Ntaryamira Day"
        if date(year, APR, 6).weekday() == SUN:
            self[date(year, APR, 7)] = name + " (Observed)"

        # Labour Day
        name = "Labour Day"
        self[date(year, MAY, 1)] = name
        if date(year, MAY, 1).weekday() == SUN:
            self[date(year, MAY, 2)] = name + " (Observed)"

        # Ascension Day
        name = "Ascension Day"
        self[easter(year) + rd(days=+39)] = name

        # Independence Day post 1962
        name = "Independence Day"
        if year > 1961:
            self[date(year, JUL, 1)] = name
            if date(year, JUL, 1).weekday() == SUN:
                self[date(year, JUL, 2)] = name + " (Observed)"

        # Eid Al Adha- Feast of the Sacrifice
        # date of observance is announced yearly
        for yr in (year - 1, year):
            for date_obs in _islamic_to_gre(yr, 12, 10):
                hol_date = date_obs
                _add_holiday(hol_date, "Eid Al Adha")
                _add_holiday(hol_date + rd(days=1), "Eid Al Adha")

        # Assumption Day
        name = "Assumption Day"
        self[date(year, AUG, 15)] = name

        # Prince Louis Rwagasore Day
        name = "Prince Louis Rwagasore Day"
        self[date(year, OCT, 13)] = name
        if date(year, OCT, 13).weekday() == SUN:
            self[date(year, OCT, 14)] = name + " (Observed)"

        # President Ndadaye's Day
        name = "President Ndadaye's Day"
        self[date(year, OCT, 21)] = name
        if date(year, OCT, 21).weekday() == SUN:
            self[date(year, OCT, 22)] = name + " (Observed)"

        # All Saints' Day
        name = "All Saints' Day"
        self[date(year, NOV, 1)] = name
        if date(year, NOV, 1).weekday() == SUN:
            self[date(year, NOV, 2)] = name + " (Observed)"

        # Christmas Day
        self[date(year, DEC, 25)] = "Christmas Day"


class BI(Burundi):
    pass


class BDI(Burundi):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, SEP, NOV, DEC, MON
from holidays.constants import TUE, THU, SUN
from holidays.holiday_base import HolidayBase


class Angola(HolidayBase):
    """
    https://www.officeholidays.com/countries/angola/
    https://www.timeanddate.com/holidays/angola/
    """

    country = "AO"

    def _populate(self, year: int) -> None:
        # Observed since 1975
        # TODO do more research on history of Angolan holidays
        if year <= 1974:
            return
        super()._populate(year)

        self[date(year, JAN, 1)] = "Ano novo"
        # Since 2018, if the following year's New Year's Day falls on a
        # Tuesday, the 31st of the current year is also a holiday.
        if year >= 2018:
            if self.observed and date(year, DEC, 31).weekday() == MON:
                self[date(year, DEC, 31)] = "Ano novo (Day off)"

        easter_date = easter(year)
        self[(easter_date + rd(days=-2))] = "Sexta-feira Santa"

        # carnival is the Tuesday before Ash Wednesday
        # which is 40 days before easter excluding sundays
        self[(easter_date + rd(days=-47))] = "Carnaval"

        self[date(year, FEB, 4)] = "Dia do Início da Luta Armada"
        self[date(year, MAR, 8)] = "Dia Internacional da Mulher"

        if year >= 2019:
            self[date(year, MAR, 23)] = "Dia da Libertação da África Austral"

        self[date(year, APR, 4)] = "Dia da Paz e Reconciliação"
        self[date(year, MAY, 1)] = "Dia Mundial do Trabalho"

        if year >= 1980:
            self[date(year, SEP, 17)] = "Dia do Herói Nacional"

        self[date(year, NOV, 2)] = "Dia dos Finados"
        self[date(year, NOV, 11)] = "Dia da Independência"
        self[date(year, DEC, 25)] = "Dia de Natal e da Família"

        # As of 1995/1/1, whenever a public holiday falls on a Sunday,
        # it rolls over to the following Monday
        # Since 2018 when a public holiday falls on the Tuesday or Thursday
        # the Monday or Friday is also a holiday
        if self.observed and year >= 1995:
            for k, v in list(self.items()):
                if k.year != year:
                    continue
                if year <= 2017:
                    if k.weekday() == SUN:
                        self[k + rd(days=+1)] = v + " (Observed)"
                else:
                    if k.weekday() == TUE and k != date(year, JAN, 1):
                        self[k + rd(days=-1)] = v + " (Day off)"
                    elif k.weekday() == THU:
                        self[k + rd(days=+1)] = v + " (Day off)"


class AO(Angola):
    pass


class AGO(Angola):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, JUL, OCT, DEC, SUN
from holidays.holiday_base import HolidayBase


class Cuba(HolidayBase):
    country = "CU"

    def _populate(self, year):
        super()._populate(year)

        """
        Overview: https://en.wikipedia.org/wiki/Public_holidays_in_Cuba
        1984 (DEC 28): https://bit.ly/3okNBbt
        2007 (NOV 19): https://bit.ly/3oFbhaZ
        2013 (DEC 20): https://bit.ly/3zoO3vC
        Note: for holidays that can be moved to a Monday if they fall on a
              Sunday, between 1984 and 2013, the State Committee of Work and
              Social Security would determine if they would be moved to the
              Monday, or if they would stay on the Sunday, presumably depending
              on quotas. After 2013, they always move to Monday. I could not
              find any records of this, so I implemented this making it always
              go to the next Monday.
        """

        name = "Aniversario de la Revolución [Anniversary of the Revolution]"
        self[date(year, JAN, 1)] = name
        if (
            year <= 2013
            and self.observed
            and date(year, JAN, 1).weekday() == SUN
        ):
            self[date(year, JAN, 1) + rd(weekday=MO)] = name + " (Observed)"

        # Granted in 2007 decree.
        if year > 2007:
            self[date(year, JAN, 2)] = "Día de la Victoria [Victory Day]"

        # Granted temporarily in 2012 and 2013:
        #   https://cnn.it/3v5V6GY
        #   https://bit.ly/3v6bM18
        # Permanently granted in 2013 decree for 2014 and onwards.
        if year >= 2012:
            self[easter(year) + rd(days=-2)] = "Viernes Santo [Good Friday]"

        name = "Día Internacional de los Trabajadores [Labour Day]"
        self[date(year, MAY, 1)] = name
        if self.observed and date(year, MAY, 1).weekday() == SUN:
            self[date(year, MAY, 1) + rd(weekday=MO)] = name + " (Observed)"

        self[date(year, JUL, 25)] = (
            "Conmemoración del asalto a Moncada "
            "[Commemoration of the Assault of the Moncada garrison]"
        )

        self[
            date(year, JUL, 26)
        ] = "Día de la Rebeldía Nacional [Day of the National Rebellion]"

        self[date(year, JUL, 27)] = (
            "Conmemoración del asalto a Moncada "
            "[Commemoration of the Assault of the Moncada garrison]"
        )

        name = "Inicio de las Guerras de Independencia [Independence Day]"
        self[date(year, OCT, 10)] = name
        if self.observed and date(year, OCT, 10).weekday() == SUN:
            self[date(year, OCT, 10) + rd(weekday=MO)] = name + " (Observed)"

        # In 1969, Christmas was cancelled for the sugar harvest but then was
        # cancelled for good:
        #   https://bit.ly/3OpwX5i
        # In 1997, Christmas was temporarily back for the pope's visit:
        #   https://cnn.it/3Omn349
        # In 1998, Christmas returns for good:
        #   https://bit.ly/3cyXhwz
        #   https://bit.ly/3cyXj7F
        if year >= 1997 or year < 1969:
            self[date(year, DEC, 25)] = "Día de Navidad [Christmas Day]"

        # Granted in 2007 decree.
        if year >= 2007:
            self[date(year, DEC, 31)] = "Fiesta de Fin de Año [New Year's Eve]"


class CU(Cuba):
    pass


class CUB(Cuba):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country and subdivision
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

__all__ = (
    "CountryHoliday",
    "country_holidays",
    "financial_holidays",
    "list_supported_countries",
    "list_supported_financial",
)

import inspect
import warnings
from datetime import date, timedelta
from functools import lru_cache
from typing import Dict, Iterable, List, Optional, Union

from hijri_converter import convert
from hijri_converter.ummalqura import GREGORIAN_RANGE

from holidays import countries, financial
from holidays.holiday_base import HolidayBase


def country_holidays(
    country: str,
    subdiv: Optional[str] = None,
    years: Optional[Union[int, Iterable[int]]] = None,
    expand: bool = True,
    observed: bool = True,
    prov: Optional[str] = None,
    state: Optional[str] = None,
) -> HolidayBase:
    """
    Returns a new dictionary-like :py:class:`HolidayBase` object for the public
    holidays of the country matching **country** and other keyword arguments.

    :param country:
        An ISO 3166-1 Alpha-2 country code.

    :param subdiv:
        The subdivision (e.g. state or province); not implemented for all
        countries (see documentation).

    :param years:
        The year(s) to pre-calculate public holidays for at instantiation.

    :param expand:
        Whether the entire year is calculated when one date from that year
        is requested.

    :param observed:
        Whether to include the dates of when public holiday are observed
        (e.g. a holiday falling on a Sunday being observed the following
        Monday). False may not work for all countries.

    :param prov:
        *deprecated* use subdiv instead.

    :param state:
        *deprecated* use subdiv instead.

    :return:
        A :py:class:`HolidayBase` object matching the **country**.

    The key of the :class:`dict`-like :class:`HolidayBase` object is the
    `date` of the holiday, and the value is the name of the holiday itself.
    Dates where a key is not present are not public holidays (or, if
    **observed** is False, days when a public holiday is observed).

    When passing the `date` as a key, the `date` can be expressed in one of the
    following types:

    * :class:`datetime.date`,
    * :class:`datetime.datetime`,
    * a :class:`str` of any format recognized by :func:`dateutil.parser.parse`,
    * or a :class:`float` or :class:`int` representing a POSIX timestamp.

    The key is always returned as a :class:`datetime.date` object.

    To maximize speed, the list of public holidays is built on the fly as
    needed, one calendar year at a time. When the object is instantiated
    without a **years** parameter, it is empty, but, unless **expand** is set
    to False, as soon as a key is accessed the class will calculate that entire
    year's list of holidays and set the keys with them.

    If you need to list the holidays as opposed to querying individual dates,
    instantiate the class with the **years** parameter.

    Example usage:

    >>> from holidays import country_holidays
    >>> us_holidays = country_holidays('US')
    # For a specific subdivision (e.g. state or province):
    >>> calif_holidays = country_holidays('US', subdiv='CA')

    The below will cause 2015 holidays to be calculated on the fly:

    >>> from datetime import date
    >>> assert date(2015, 1, 1) in us_holidays

    This will be faster because 2015 holidays are already calculated:

    >>> assert date(2015, 1, 2) not in us_holidays

    The :class:`HolidayBase` class also recognizes strings of many formats
    and numbers representing a POSIX timestamp:

    >>> assert '2014-01-01' in us_holidays
    >>> assert '1/1/2014' in us_holidays
    >>> assert 1388597445 in us_holidays

    Show the holiday's name:

    >>> us_holidays.get('2014-01-01')
    "New Year's Day"

    Check a range:

    >>> us_holidays['2014-01-01': '2014-01-03']
    [datetime.date(2014, 1, 1)]

    List all 2020 holidays:

    >>> us_holidays = country_holidays('US', years=2020)
    >>> for day in us_holidays.items():
    ...     print(day)
    (datetime.date(2020, 1, 1), "New Year's Day")
    (datetime.date(2020, 1, 20), 'Martin Luther King Jr. Day')
    (datetime.date(2020, 2, 17), "Washington's Birthday")
    (datetime.date(2020, 5, 25), 'Memorial Day')
    (datetime.date(2020, 7, 4), 'Independence Day')
    (datetime.date(2020, 7, 3), 'Independence Day (Observed)')
    (datetime.date(2020, 9, 7), 'Labor Day')
    (datetime.date(2020, 10, 12), 'Columbus Day')
    (datetime.date(2020, 11, 11), 'Veterans Day')
    (datetime.date(2020, 11, 26), 'Thanksgiving')
    (datetime.date(2020, 12, 25), 'Christmas Day')

    Some holidays are only present in parts of a country:

    >>> us_pr_holidays = country_holidays('US', subdiv='PR')
    >>> assert '2018-01-06' not in us_holidays
    >>> assert '2018-01-06' in us_pr_holidays

    Append custom holiday dates by passing one of:

    * a :class:`dict` with date/name key/value pairs (e.g.
      ``{'2010-07-10': 'My birthday!'}``),
    * a list of dates (as a :class:`datetime.date`, :class:`datetime.datetime`,
      :class:`str`, :class:`int`, or :class:`float`); ``'Holiday'`` will be
      used as a description,
    * or a single date item (of one of the types above); ``'Holiday'`` will be
      used as a description:

    >>> custom_holidays = country_holidays('US', years=2015)
    >>> custom_holidays.update({'2015-01-01': "New Year's Day"})
    >>> custom_holidays.update(['2015-07-01', '07/04/2015'])
    >>> custom_holidays.update(date(2015, 12, 25))
    >>> assert date(2015, 1, 1) in custom_holidays
    >>> assert date(2015, 1, 2) not in custom_holidays
    >>> assert '12/25/2015' in custom_holidays

    For more complex logic, like 4th Monday of January, you can inherit the
    :class:`HolidayBase` class and define your own :meth:`_populate` method.
    See documentation for examples.
    """
    try:
        country_classes = inspect.getmembers(countries, inspect.isclass)
        country_class = next(
            obj for name, obj in country_classes if name == country
        )
        country_holiday: HolidayBase = country_class(
            years=years,
            subdiv=subdiv,
            expand=expand,
            observed=observed,
            prov=prov,
            state=state,
        )
    except StopIteration:
        raise NotImplementedError(f"Country {country} not available")
    return country_holiday


def financial_holidays(
    market: str,
    subdiv: Optional[str] = None,
    years: Optional[Union[int, Iterable[int]]] = None,
    expand: bool = True,
    observed: bool = True,
) -> HolidayBase:
    """
    Returns a new dictionary-like :py:class:`HolidayBase` object for the public
    holidays of the financial market matching **market** and other keyword
    arguments.

    :param market:
        An ISO 3166-1 Alpha-2 market code.

    :param subdiv:
        Currently not implemented for markets (see documentation).

    :param years:
        The year(s) to pre-calculate public holidays for at instantiation.

    :param expand:
        Whether the entire year is calculated when one date from that year
        is requested.

    :param observed:
        Whether to include the dates of when public holiday are observed
        (e.g. a holiday falling on a Sunday being observed the following
        Monday). False may not work for all countries.

    :return:
        A :py:class:`HolidayBase` object matching the **market**.

    Example usage:

    >>> from holidays import financial_holidays
    >>> nyse_holidays = financial_holidays('NYSE')

    See :py:func:`country_holidays` documentation for further details and
    examples.
    """
    try:
        financial_classes = inspect.getmembers(financial, inspect.isclass)
        financial_class = next(
            obj for name, obj in financial_classes if name == market
        )
        financial_holiday: HolidayBase = financial_class(
            years=years,
            subdiv=subdiv,
            expand=expand,
            observed=observed,
        )
    except StopIteration:
        raise NotImplementedError(f"Financial market {market} not available")
    return financial_holiday


def CountryHoliday(
    country: str,
    subdiv: Optional[str] = None,
    years: Optional[Union[int, Iterable[int]]] = None,
    expand: bool = True,
    observed: bool = True,
    prov: Optional[str] = None,
    state: Optional[str] = None,
) -> HolidayBase:
    """
    Deprecated name for :py:func:`country_holidays`.

    :meta private:
    """

    warnings.warn(
        "CountryHoliday is deprecated, use country_holidays instead.",
        DeprecationWarning,
    )
    return country_holidays(
        country, subdiv, years, expand, observed, prov, state
    )


def list_supported_countries() -> Dict[str, List[str]]:
    """
    Get all supported countries and their subdivisions.

    :return:
        A dictionary where the key is the ISO 3166-1 Alpha-2 country codes and
        the value is a list of supported subdivision codes.
    """
    return {
        cls.country: cls.subdivisions
        for name, cls in inspect.getmembers(countries, inspect.isclass)
        if len(name) == 2 and issubclass(cls, HolidayBase)
    }


def list_supported_financial() -> Dict[str, List[str]]:
    """
    Get all supported financial markets and their subdivisions.

    :return:
        A dictionary where the key is the market codes and
        the value is a list of supported subdivision codes.
    """
    return {
        cls.market: cls.subdivisions
        for _, cls in inspect.getmembers(financial, inspect.isclass)
        if issubclass(cls, HolidayBase)
    }


def _islamic_to_gre(Gyear: int, Hmonth: int, Hday: int) -> List[date]:
    """
    Find the Gregorian dates of all instances of Islamic (Lunar Hijrī) calendar
    month and day falling within the Gregorian year. There could be up to two
    such instances in a single Gregorian year since the Islamic (Lunar Hijrī)
    calendar is about 11 days shorter.

    Relies on package `hijri_converter
    <https://www.pypy.org/package/hijri_converter>`__.

    :param Gyear:
        The Gregorian year.

    :param Hmonth:
        The Lunar Hijrī (Islamic) month.

    :param Hday:
        The Lunar Hijrī (Islamic) day.

    :return:
        List of Gregorian dates within the Gregorian year specified that
        matches the Islamic (Lunar Hijrī) calendar day and month specified.
    """
    gre_dates: List[date] = []

    # To avoid hijri_converter check range OverflowError.
    dt = (Gyear, Hmonth, Hday)
    dt_min, dt_max = GREGORIAN_RANGE
    if dt < dt_min or dt > dt_max:
        return gre_dates

    Hyear = convert.Gregorian(Gyear, 1, 1).to_hijri().datetuple()[0]
    gres = [
        convert.Hijri(y, Hmonth, Hday).to_gregorian()
        for y in range(Hyear - 1, Hyear + 2)
    ]
    gre_dates.extend(
        (date(*gre.datetuple()) for gre in gres if gre.year == Gyear)
    )

    return gre_dates


class _ChineseLuniSolar:
    def __init__(self) -> None:
        """
        This class has functions that generate Gregorian dates for holidays
        based on the Chinese lunisolar calendar.

        See `Wikipedia
        <https://en.wikipedia.org/wiki/Chinese_New_Year#Dates_in_Chinese_\
        lunisolar_calendar>`__

        Usage example:

        >>> from holidays.utils import _ChineseLuniSolar
        >>> cnls = _ChineseLuniSolar()
        >>> print(cnls.lunar_n_y_date(2010))
        2010-02-14
        """

        # A binary representation starting from year 1901 of the number of
        # days per year, and the number of days from the 1st to the 13th to
        # store the monthly (including the month of the month). 1 means that
        # the month is 30 days. 0 means the month is 29 days.
        # The 12th to 15th digits indicate the month of the next month.
        # If it is 0x0F, it means that there is no leap month.
        self.G_LUNAR_MONTH_DAYS = [
            0xF0EA4,  # 1901
            0xF1D4A,
            0x52C94,
            0xF0C96,
            0xF1536,
            0x42AAC,
            0xF0AD4,
            0xF16B2,
            0x22EA4,
            0xF0EA4,  # 1911
            0x6364A,
            0xF164A,
            0xF1496,
            0x52956,
            0xF055A,
            0xF0AD6,
            0x216D2,
            0xF1B52,
            0x73B24,
            0xF1D24,  # 1921
            0xF1A4A,
            0x5349A,
            0xF14AC,
            0xF056C,
            0x42B6A,
            0xF0DA8,
            0xF1D52,
            0x23D24,
            0xF1D24,
            0x61A4C,  # 1931
            0xF0A56,
            0xF14AE,
            0x5256C,
            0xF16B4,
            0xF0DA8,
            0x31D92,
            0xF0E92,
            0x72D26,
            0xF1526,
            0xF0A56,  # 1941
            0x614B6,
            0xF155A,
            0xF0AD4,
            0x436AA,
            0xF1748,
            0xF1692,
            0x23526,
            0xF152A,
            0x72A5A,
            0xF0A6C,  # 1951
            0xF155A,
            0x52B54,
            0xF0B64,
            0xF1B4A,
            0x33A94,
            0xF1A94,
            0x8152A,
            0xF152E,
            0xF0AAC,
            0x6156A,  # 1961
            0xF15AA,
            0xF0DA4,
            0x41D4A,
            0xF1D4A,
            0xF0C94,
            0x3192E,
            0xF1536,
            0x72AB4,
            0xF0AD4,
            0xF16D2,  # 1971
            0x52EA4,
            0xF16A4,
            0xF164A,
            0x42C96,
            0xF1496,
            0x82956,
            0xF055A,
            0xF0ADA,
            0x616D2,
            0xF1B52,  # 1981
            0xF1B24,
            0x43A4A,
            0xF1A4A,
            0xA349A,
            0xF14AC,
            0xF056C,
            0x60B6A,
            0xF0DAA,
            0xF1D92,
            0x53D24,  # 1991
            0xF1D24,
            0xF1A4C,
            0x314AC,
            0xF14AE,
            0x829AC,
            0xF06B4,
            0xF0DAA,
            0x52D92,
            0xF0E92,
            0xF0D26,  # 2001
            0x42A56,
            0xF0A56,
            0xF14B6,
            0x22AB4,
            0xF0AD4,
            0x736AA,
            0xF1748,
            0xF1692,
            0x53526,
            0xF152A,  # 2011
            0xF0A5A,
            0x4155A,
            0xF156A,
            0x92B54,
            0xF0BA4,
            0xF1B4A,
            0x63A94,
            0xF1A94,
            0xF192A,
            0x42A5C,  # 2021
            0xF0AAC,
            0xF156A,
            0x22B64,
            0xF0DA4,
            0x61D52,
            0xF0E4A,
            0xF0C96,
            0x5192E,
            0xF1956,
            0xF0AB4,  # 2031
            0x315AC,
            0xF16D2,
            0xB2EA4,
            0xF16A4,
            0xF164A,
            0x63496,
            0xF1496,
            0xF0956,
            0x50AB6,
            0xF0B5A,  # 2041
            0xF16D4,
            0x236A4,
            0xF1B24,
            0x73A4A,
            0xF1A4A,
            0xF14AA,
            0x5295A,
            0xF096C,
            0xF0B6A,
            0x31B54,  # 2051
            0xF1D92,
            0x83D24,
            0xF1D24,
            0xF1A4C,
            0x614AC,
            0xF14AE,
            0xF09AC,
            0x40DAA,
            0xF0EAA,
            0xF0E92,  # 2061
            0x31D26,
            0xF0D26,
            0x72A56,
            0xF0A56,
            0xF14B6,
            0x52AB4,
            0xF0AD4,
            0xF16CA,
            0x42E94,
            0xF1694,  # 2071
            0x8352A,
            0xF152A,
            0xF0A5A,
            0x6155A,
            0xF156A,
            0xF0B54,
            0x4174A,
            0xF1B4A,
            0xF1A94,
            0x3392A,  # 2081
            0xF192C,
            0x7329C,
            0xF0AAC,
            0xF156A,
            0x52B64,
            0xF0DA4,
            0xF1D4A,
            0x41C94,
            0xF0C96,
            0x8192E,  # 2091
            0xF0956,
            0xF0AB6,
            0x615AC,
            0xF16D4,
            0xF0EA4,
            0x42E4A,
            0xF164A,
            0xF1516,
            0x22936,  # 2100
        ]
        # Define range of years covered
        self.START_YEAR = 1901
        self.END_YEAR = 2099
        # The 1st day of the 1st month of the Gregorian calendar is 1901/2/19
        self.LUNAR_START_DATE = ((1901, 1, 1),)
        self.SOLAR_START_DATE = date(1901, 2, 19)
        # The Gregorian date for December 30, 2099 is 2100/2/8
        self.LUNAR_END_DATE = (2099, 12, 30)
        self.SOLAR_END_DATE = date(2100, 2, 18)

    @lru_cache()
    def _get_leap_month(self, lunar_year: int) -> int:
        """
        Calculate the leap lunar month in a lunar year.

        :param lunar_year:
            The lunar year.

        :return:
            The number of the leap month if one exists in the year, otherwise
            15.
        """
        return (
            self.G_LUNAR_MONTH_DAYS[lunar_year - self.START_YEAR] >> 16
        ) & 0x0F

    def _lunar_month_days(self, lunar_year: int, lunar_month: int) -> int:
        """
        Calculate the number of days in a lunar month.

        :param lunar_year:
            The lunar year.

        :param lunar_month:
            The lunar month of the lunar year.

        :return:
            The number of days in the lunar month.
        """
        return 29 + (
            (
                self.G_LUNAR_MONTH_DAYS[lunar_year - self.START_YEAR]
                >> lunar_month
            )
            & 0x01
        )

    def _lunar_year_days(self, year: int) -> int:
        """
        Calculate the number of days in a lunar year.

        :param year:
            The lunar year.

        :return:
            The number of days in the lunar year.
        """
        days = 0
        months_day = self.G_LUNAR_MONTH_DAYS[year - self.START_YEAR]
        for i in range(1, 13 if self._get_leap_month(year) == 0x0F else 14):
            day = 29 + ((months_day >> i) & 0x01)
            days += day
        return days

    @lru_cache()
    def _span_days(self, year: int) -> int:
        """
        Calculate the number of days elapsed since self.SOLAR_START_DATE to the
        beginning of the year.

        :param year:
            The year.

        :return:
             The number of days since self.SOLAR_START_DATE.
        """
        span_days = 0
        for y in range(self.START_YEAR, year):
            span_days += self._lunar_year_days(y)
        return span_days

    def lunar_n_y_date(self, year: int) -> date:
        """
        Calculate the Gregorian date of Chinese Lunar New Year.

        This is a faster implementation than calling
        ``lunar_to_gre(year, 1, 1)``.

        :param year:
            The Gregorian year.

        :return:
            The Gregorian date of Chinese Lunar New Year.
        """
        # The Chinese calendar defines the lunar month containing the winter
        # solstice as the eleventh month, which means that Chinese New Year
        # usually falls on the second new moon after the winter solstice
        # (rarely the third if an intercalary month intervenes). In more
        # than 96 percent of the years, Chinese New Year's Day is the closest
        # date to a new moon to lichun (Chinese: 立春; "start of spring") on 4
        # or 5 February, and the first new moon after dahan (Chinese: 大寒;
        # "major cold"). In the Gregorian calendar, the Chinese New Year begins
        # at the new moon that falls between 21 January and 20 February.

        span_days = self._span_days(year)
        # Always in first month (by definition)
        # leap_month = self._get_leap_month(year)
        # for m in range(1, 1 + (1 > leap_month)):
        #     span_days += self._lunar_month_days(year, m)
        return self.SOLAR_START_DATE + timedelta(span_days)

    def lunar_to_gre(
        self, year: int, month: int, day: int, leap: bool = True
    ) -> date:
        """
        Calculate the Gregorian date of a Chinese lunar day and month in a
        given Gregorian year.

        :param year:
            The Gregorian year.

        :param year:
            The Chinese lunar month.

        :param year:
            The Chinese lunar day.

        :return:
            The Gregorian date.
        """
        span_days = self._span_days(year)
        leap_month = self._get_leap_month(year) if leap else 15
        for m in range(1, month + (month > leap_month)):
            span_days += self._lunar_month_days(year, m)
        span_days += day - 1
        return self.SOLAR_START_DATE + timedelta(span_days)

    def vesak_date(self, year: int) -> date:
        """
        Calculate the estimated Gregorian date of Vesak for Thailand, Laos,
        Singapore and Indonesia, corresponding to the fourteenth day of the
        fourth month in the Chinese lunar calendar. See `Wikipedia
        <https://en.wikipedia.org/wiki/Vesak#Dates_of_observance>`__.

        :param year:
            The Gregorian year.

        :return:
            Estimated Gregorian date of Vesak (14th day of 4th month of the
            lunar calendar).
        """
        span_days = self._span_days(year)
        leap_month = self._get_leap_month(year)
        for m in range(1, 4 + (4 > leap_month)):
            span_days += self._lunar_month_days(year, m)
        span_days += 14
        return self.SOLAR_START_DATE + timedelta(span_days)

    def vesak_may_date(self, year: int) -> date:
        """
        Calculate the estimated Gregorian date of Vesak for Sri Lanka, Nepal,
        India, Bangladesh and Malaysia, corresponding to the day of the
        first full moon in May in the Gregorian calendar. See `Wikipedia
        <https://en.wikipedia.org/wiki/Vesak#Dates_of_observance>`__.

        :param year:
            The Gregorian year.

        :return:
            Estimated Gregorian date of Vesak (first full moon in May).
        """
        span_days = self._span_days(year)
        vesak_may_date = self.SOLAR_START_DATE + timedelta(span_days + 14)
        m = 1
        while vesak_may_date.month < 5:
            vesak_may_date += timedelta(self._lunar_month_days(year, m))
            m += 1
        return vesak_may_date

    def s_diwali_date(self, year: int) -> date:
        """
        Calculate the estimated Gregorian date of Southern India (Tamil)
        Diwali.

        Defined as the date of Amāvásyā (new moon) of Kārttikai, which
        corresponds with the months of November or December in the Gregorian
        calendar. See `Wikipedia <https://en.wikipedia.org/wiki/Diwali>`__.

        :param year:
            The Gregorian year.

        :return:
            Estimated Gregorian date of Southern India (Tamil) Diwali.
        """
        span_days = self._span_days(year)
        leap_month = self._get_leap_month(year)
        for m in range(1, 10 + (10 > leap_month)):
            span_days += self._lunar_month_days(year, m)
        span_days -= 2
        return self.SOLAR_START_DATE + timedelta(span_days)

    def thaipusam_date(self, year: int) -> date:
        """
        Calculate the estimated Gregorian date of Thaipusam (Tamil).

        Defined as the date of the full moon in the Tamil month of Thai, which
        corresponds with the months of January or February in the Gregorian
        calendar. See `Wikipedia <https://en.wikipedia.org/wiki/Thaipusam>`__.

        :param year:
            The Gregorian year.

        :return:
            Estimated Gregorian date of Thaipusam (Tamil).
        """
        span_days = self._span_days(year)
        leap_month = self._get_leap_month(year)
        for m in range(1, 1 + (leap_month <= 6)):
            span_days += self._lunar_month_days(year, m)
        span_days -= 15
        return self.SOLAR_START_DATE + timedelta(span_days)



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import SU
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, MAY, JUN, AUG, NOV, DEC
from holidays.holiday_base import HolidayBase


class Madagascar(HolidayBase):
    """
    https://www.officeholidays.com/countries/madagascar
    https://www.timeanddate.com/holidays/madagascar/
    """

    country = "MG"

    def _populate(self, year):
        # Observed since 1947
        if year <= 1946:
            return
        super()._populate(year)

        self[date(year, JAN, 1)] = "Taom-baovao / New Year's Day"
        self[date(year, MAR, 8)] = "Fetin'ny vehivavy / Women's Day"
        self[date(year, MAR, 29)] = "Fetin'ny mahery fo / Martyrs' Day"
        self[date(year, MAY, 1)] = "Labour Day"
        self[
            date(year, JUN, 1) + rd(weekday=SU(+3))
        ] = "Fetin'ny ray / Father's Day"

        if year >= 1960:
            self[date(year, JUN, 26)] = "Independence Day"

        self[
            date(year, AUG, 15)
        ] = "Fiakaran'ny Masina Maria tany an-danitra / Assumption Day"

        self[date(year, NOV, 1)] = "Fetin'ny olo-masina / All Saints' Day"

        if year >= 2011:
            self[date(year, DEC, 11)] = "Republic Day"

        self[date(year, DEC, 25)] = "Fetin'ny noely / Christmas Day"

        easter_date = easter(year)
        self[easter_date] = "Fetin'ny paska / Easter Sunday"
        self[easter_date + rd(days=+1)] = "Alatsinain'ny paska / Easter Monday"
        self[
            easter_date + rd(days=+39)
        ] = "Fiakaran'ny Jesosy kristy tany an-danitra / Ascension Day"

        whit_sunday = easter_date + rd(days=+49)
        self[whit_sunday] = "Pentekosta / Whit Sunday"

        self[
            easter_date + rd(days=+50)
        ] = "Alatsinain'ny pentekosta / Whit Monday"

        dt = date(year, MAY, 31) + rd(weekday=SU(-1))
        if dt == whit_sunday:
            dt += rd(days=+7)
        self[dt] = "Fetin'ny Reny / Mother's Day"


class MG(Madagascar):
    pass


class MDG(Madagascar):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import EASTER_ORTHODOX, easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, JUN, AUG, NOV, DEC
from holidays.holiday_base import HolidayBase


class Romania(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Romania
    """

    country = "RO"

    def _populate(self, year):
        super()._populate(year)

        easter_date = easter(year, method=EASTER_ORTHODOX)

        # New Year
        for day in (1, 2):
            self[date(year, JAN, day)] = "Anul Nou"

        # Anniversary of the formation of the United Principalities
        self[date(year, JAN, 24)] = "Unirea Principatelor Române"

        # Easter (Friday, Sunday and Monday)
        for day_after_easter in (-2, 0, 1):
            self[easter_date + rd(days=day_after_easter)] = "Paștele"

        # Labour Day
        self[date(year, MAY, 1)] = "Ziua Muncii"

        # Children's Day (since 2017)
        if year >= 2017:
            self[date(year, JUN, 1)] = "Ziua Copilului"

        # Whit Monday
        for day_after_easter in (49, 50):
            self[easter_date + rd(days=day_after_easter)] = "Rusaliile"

        # Assumption of Mary
        self[date(year, AUG, 15)] = "Adormirea Maicii Domnului"

        # St. Andrew's Day
        self[date(year, NOV, 30)] = "Sfântul Andrei"

        # National Day
        self[date(year, DEC, 1)] = "Ziua Națională a României"

        # Christmas Day
        self[date(year, DEC, 25)] = "Crăciunul"
        self[date(year, DEC, 26)] = "Crăciunul"


class RO(Romania):
    pass


class ROU(Romania):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.utils import _ChineseLuniSolar, _islamic_to_gre


class Indonesia(HolidayBase):
    country = "ID"
    special_holidays = {
        # Election Day
        2018: ((JUN, 27, "Hari Pemilihan"),),
        2019: ((APR, 17, "Hari Pemilihan"),),
        2020: ((DEC, 9, "Hari Pemilihan"),),
    }

    # https://en.wikipedia.org/wiki/Public_holidays_in_Indonesia
    # https://www.timeanddate.com/holidays/indonesia
    # https://www.officeholidays.com/countries/indonesia
    def __init__(self, **kwargs):
        self.cnls = _ChineseLuniSolar()
        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self[date(year, JAN, 1)] = "Tahun Baru Masehi"

        # Chinese New Year
        if year >= 2003:
            hol_date = self.cnls.lunar_n_y_date(year)
            self[hol_date] = "Tahun Baru Imlek"

        # Day of Silence
        # https://en.wikipedia.org/wiki/Nyepi
        dates_obs = {
            2009: (MAR, 26),
            2010: (MAR, 16),
            2011: (MAR, 5),
            2012: (MAR, 23),
            2013: (MAR, 12),
            2014: (MAR, 31),
            2015: (MAR, 21),
            2016: (MAR, 9),
            2017: (MAR, 28),
            2018: (MAR, 17),
            2019: (MAR, 7),
            2020: (MAR, 25),
            2021: (MAR, 14),
            2022: (MAR, 3),
            2023: (MAR, 22),
            2024: (MAR, 11),
            2025: (MAR, 29),
            2026: (MAR, 19),
        }
        if year in dates_obs:
            hol_date = date(year, *dates_obs[year])
            self[hol_date] = "Hari Suci Nyepi"

        # Eid al-Fitr
        dates_obs = {
            2001: ((DEC, 16),),
            2002: ((DEC, 6),),
            2003: ((NOV, 25),),
            2004: ((NOV, 14),),
            2005: ((NOV, 3),),
            2006: ((OCT, 24),),
            2007: ((OCT, 13),),
            2008: ((OCT, 1),),
            2009: ((SEP, 20),),
            2010: ((SEP, 10),),
            2011: ((AUG, 30),),
            2012: ((AUG, 19),),
            2013: ((AUG, 8),),
            2014: ((JUL, 28),),
            2015: ((JUL, 17),),
            2016: ((JUL, 6),),
            2017: ((JUN, 25),),
            2018: ((JUN, 15),),
            2019: ((JUN, 5),),
            2020: ((MAY, 24),),
            2021: ((MAY, 13),),
            2022: ((MAY, 2),),
            2023: ((APR, 22),),
        }
        if year in dates_obs:
            for date_obs in dates_obs[year]:
                hol_date = date(year, *date_obs)
                self[hol_date] = "Hari Raya Idul Fitri"
                self[
                    hol_date + rd(days=+1)
                ] = "Hari kedua dari Hari Raya Idul Fitri"
        else:
            for date_obs in _islamic_to_gre(year, 10, 1):
                hol_date = date_obs
                self[hol_date] = "Hari Raya Idul Fitri* (*estimated)"
                self[
                    hol_date + rd(days=+1)
                ] = "Hari kedua dari Hari Raya Idul Fitri* (*estimated)"

        # Eid al-Adha
        dates_obs = {
            2001: ((MAR, 6),),
            2002: ((FEB, 23),),
            2003: ((FEB, 12),),
            2004: ((FEB, 2),),
            2005: ((JAN, 21),),
            2006: ((JAN, 10), (DEC, 31)),
            2007: ((DEC, 20),),
            2008: ((DEC, 8),),
            2009: ((NOV, 27),),
            2010: ((NOV, 17),),
            2011: ((NOV, 6),),
            2012: ((OCT, 26),),
            2013: ((OCT, 15),),
            2014: ((OCT, 5),),
            2015: ((SEP, 24),),
            2016: ((SEP, 12),),
            2017: ((SEP, 1),),
            2018: ((AUG, 22),),
            2019: ((AUG, 11),),
            2020: ((JUL, 31),),
            2021: ((JUL, 20),),
            2022: ((JUL, 10),),
            2023: ((JUN, 29),),
        }
        if year in dates_obs:
            for date_obs in dates_obs[year]:
                hol_date = date(year, *date_obs)
                self[hol_date] = "Hari Raya Idul Adha"
        else:
            for date_obs in _islamic_to_gre(year, 12, 10):
                hol_date = date_obs
                self[hol_date] = "Hari Raya Idul Adha* (*estimated)"

        # Islamic New Year
        dates_obs = {
            2001: ((MAR, 26),),
            2002: ((MAR, 15),),
            2003: ((MAR, 5),),
            2004: ((FEB, 22),),
            2005: ((FEB, 10),),
            2006: ((JAN, 31),),
            2007: ((JAN, 20),),
            2008: ((JAN, 10), (DEC, 29)),
            2009: ((DEC, 18),),
            2010: ((DEC, 7),),
            2011: ((NOV, 27),),
            2012: ((NOV, 15),),
            2013: ((NOV, 5),),
            2014: ((OCT, 25),),
            2015: ((OCT, 14),),
            2016: ((OCT, 2),),
            2017: ((SEP, 21),),
            2018: ((SEP, 11),),
            2019: ((SEP, 1),),
            2020: ((AUG, 20),),
            2021: ((AUG, 11),),
            2022: ((JUL, 30),),
        }
        if year in dates_obs:
            for date_obs in dates_obs[year]:
                hol_date = date(year, *date_obs)
                self[hol_date] = "Tahun Baru Islam"
        else:
            for date_obs in _islamic_to_gre(year, 1, 1):
                hol_date = date_obs
                self[hol_date] = "Tahun Baru Islam* (*estimated)"

        # The Prophet's Birthday
        dates_obs = {
            2006: ((APR, 10),),
            2007: ((MAR, 31),),
            2008: ((MAR, 20),),
            2009: ((MAR, 9),),
            2010: ((FEB, 26),),
            2011: ((FEB, 15),),
            2012: ((FEB, 5),),
            2013: ((JAN, 24),),
            2014: ((JAN, 14),),
            2015: ((JAN, 3), (DEC, 24)),
            2016: ((DEC, 12),),
            2017: ((DEC, 1),),
            2018: ((NOV, 20),),
            2019: ((NOV, 9),),
            2020: ((OCT, 29),),
            2021: ((OCT, 19),),
            2022: ((OCT, 8),),
        }
        if year in dates_obs:
            for date_obs in dates_obs[year]:
                hol_date = date(year, *date_obs)
                self[hol_date] = "Maulid Nabi Muhammad"
        else:
            for date_obs in _islamic_to_gre(year, 3, 12):
                hol_date = date_obs
                self[hol_date] = "Maulid Nabi Muhammad* (*estimated)"

        # The Prophet's Ascension
        dates_obs = {
            2001: ((OCT, 15),),
            2002: ((OCT, 4),),
            2003: ((SEP, 24),),
            2004: ((SEP, 12),),
            2005: ((SEP, 1),),
            2006: ((AUG, 22),),
            2007: ((AUG, 11),),
            2008: ((JUL, 31),),
            2009: ((JUL, 20),),
            2010: ((JUL, 9),),
            2011: ((JUN, 29),),
            2012: ((JUN, 17),),
            2013: ((JUN, 6),),
            2014: ((MAY, 27),),
            2015: ((MAY, 16),),
            2016: ((MAY, 6),),
            2017: ((APR, 24),),
            2018: ((APR, 14),),
            2019: ((APR, 3),),
            2020: ((MAR, 22),),
            2021: ((MAR, 11),),
            2022: ((FEB, 28),),
        }
        if year in dates_obs:
            for date_obs in dates_obs[year]:
                hol_date = date(year, *date_obs)
                self[hol_date] = "Isra' Mi'raj Nabi Muhammad"
        else:
            for date_obs in _islamic_to_gre(year, 7, 27):
                hol_date = date_obs
                self[hol_date] = "Isra' Mi'raj Nabi Muhammad* (*estimated)"

        # Good Friday
        self[easter(year) + rd(days=-2)] = "Wafat Yesus Kristus"

        # Buddha's Birthday
        if year >= 1983:
            date_obs = {
                2007: (JUN, 1),
                2008: (MAY, 20),
                2009: (MAY, 9),
                2010: (MAY, 28),
                2011: (MAY, 17),
                2012: (MAY, 6),
                2013: (MAY, 25),
                2014: (MAY, 15),
                2015: (JUN, 2),
                2016: (MAY, 22),
                2017: (MAY, 11),
                2018: (MAY, 29),
                2019: (MAY, 19),
                2020: (MAY, 7),
                2021: (MAY, 26),
                2022: (MAY, 16),
                2023: (JUN, 4),
            }
            if year in date_obs:
                buddha_date = date(year, *date_obs[year])
                self[buddha_date] = "Hari Raya Waisak"
            else:
                buddha_date = self.cnls.vesak_date(year)
                self[buddha_date] = "Hari Raya Waisak* (*estimated)"

        # Labour Day
        if 1953 <= year <= 1968 or year >= 2014:
            self[date(year, MAY, 1)] = "Hari Buruh Internasional"

        # Ascension Day
        self[easter(year) + rd(days=+39)] = "Kenaikan Yesus Kristus"

        # Pancasila Day
        if year >= 2017:
            self[date(year, JUN, 1)] = "Hari Lahir Pancasila"

        # National Day
        self[date(year, AUG, 17)] = "Hari Kemerdekaan Republik Indonesia"

        # Christmas Day
        self[date(year, DEC, 25)] = "Hari Raya Natal"


class ID(Indonesia):
    pass


class IDN(Indonesia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, MAY, AUG, SEP, DEC, SUN
from holidays.holiday_base import HolidayBase


class Namibia(HolidayBase):
    """
    https://www.officeholidays.com/countries/namibia
    https://www.timeanddate.com/holidays/namibia/

    """

    country = "NA"
    special_holidays = {
        # https://gazettes.africa/archive/na/1999/na-government-gazette-dated-1999-11-22-no-2234.pdf
        1999: ((DEC, 31, "Y2K changeover"),),
        2000: ((JAN, 3, "Y2K changeover"),),
    }

    def _populate(self, year):
        if year <= 1989:
            return
        super()._populate(year)

        self[date(year, JAN, 1)] = "New Year's Day"
        self[date(year, MAR, 21)] = "Independence Day"

        # Easter Calculation
        easter_date = easter(year)
        self[easter_date + rd(days=-2)] = "Good Friday"
        self[easter_date + rd(days=+1)] = "Easter Monday"
        self[easter_date + rd(days=+39)] = "Ascension Day"
        # --------END OF EASTER------------#

        self[date(year, MAY, 1)] = "Workers' Day"
        self[date(year, MAY, 4)] = "Cassinga Day"
        self[date(year, MAY, 25)] = "Africa Day"
        self[date(year, AUG, 26)] = "Heroes' Day"

        dt = date(year, SEP, 10)
        if year >= 2005:
            # http://www.lac.org.na/laws/2004/3348.pdf
            self[dt] = "Day of the Namibian Women and Intr. Human Rights Day"
        else:
            self[dt] = "International Human Rights Day"

        self[date(year, DEC, 25)] = "Christmas Day"
        self[date(year, DEC, 26)] = "Family Day"

        # https://tinyurl.com/lacorg5835
        # As of 1991/2/1, whenever a public holiday falls on a Sunday,
        # it rolls over to the monday, unless that monday is already
        # a public holiday.
        if self.observed:
            for k, v in list(self.items()):
                if k.weekday() == SUN and k.year == year:
                    self[k + rd(days=+1)] = v + " (Observed)"


class NA(Namibia):
    pass


class NAM(Namibia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, JUN, FRI, SAT
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre

# Since Djibouti share most of it's holidays with other muslim countries,
# this class is just a copy of Egypt's.


class Djibouti(HolidayBase):

    # Holidays here are estimates, it is common for the day to be pushed
    # if falls in a weekend, although not a rule that can be implemented.
    # Holidays after 2020: the following four moving date holidays whose exact
    # date is announced yearly are estimated (and so denoted):
    # - Eid El Fetr*
    # - Eid El Adha*
    # - Isra wal Miraj*
    # - Moulad El Naby*
    # - Arafat*
    # *only if hijri-converter library is installed, otherwise a warning is
    #  raised that this holiday is missing. hijri-converter requires
    #  Python >= 3.6
    # is_weekend function is there, however not activated for accuracy.

    country = "DJ"
    weekend = {FRI, SAT}

    def _populate(self, year):
        super()._populate(year)

        def _add_holiday(dt: date, hol: str) -> None:
            """Only add if in current year; prevents adding holidays across
            years (handles multi-day Islamic holidays that straddle Gregorian
            years).
            """
            if dt.year == year:
                self[dt] = hol

        # New Year's Day
        self[date(year, JAN, 1)] = "Nouvel an"

        # Labour Day
        self[date(year, MAY, 1)] = "Fête du travail"

        # Fête de l'indépendance
        self[date(year, JUN, 27)] = "Fête de l'indépendance"
        self[date(year, JUN, 28)] = "Fête de l'indépendance"

        # Isra wal Miraj
        # The night journey of the prophet Muhammad
        for date_obs in _islamic_to_gre(year, 7, 27):
            hol_date = date_obs
            self[hol_date] = "Isra wal Miraj"

        # Eid al-Fitr - Feast Festive
        # date of observance is announced yearly, This is an estimate since
        # having the Holiday on Weekend does change the number of days,
        # deceided to leave it since marking a Weekend as a holiday
        # wouldn't do much harm.
        for yr in (year - 1, year):
            for date_obs in _islamic_to_gre(yr, 10, 1):
                hol_date = date_obs
                _add_holiday(hol_date, "Eid al-Fitr")
                _add_holiday(
                    hol_date + rd(days=1), "Eid al-Fitr deuxième jour"
                )

        # Arafat & Eid al-Adha - Scarfice Festive
        # date of observance is announced yearly
        for yr in (year - 1, year):
            for date_obs in _islamic_to_gre(yr, 12, 9):
                hol_date = date_obs
                _add_holiday(hol_date, "Arafat")
                _add_holiday(hol_date + rd(days=1), "Eid al-Adha")
                _add_holiday(
                    hol_date + rd(days=2), "Eid al-Adha deuxième jour"
                )

        # Islamic New Year - (hijari_year, 1, 1)
        for date_obs in _islamic_to_gre(year, 1, 1):
            hol_date = date_obs
            self[hol_date] = "Nouvel an musulman"

        # Prophet Muhammad's Birthday - (hijari_year, 3, 12)
        for date_obs in _islamic_to_gre(year, 3, 12):
            hol_date = date_obs
            self[hol_date] = "Naissance du prophet Muhammad"


class DJ(Djibouti):
    pass


class DJI(Djibouti):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date, datetime

from dateutil import tz
from dateutil.easter import easter
from dateutil.relativedelta import MO
from dateutil.relativedelta import relativedelta as rd
from pymeeus.Epoch import Epoch
from pymeeus.Sun import Sun

from holidays.constants import JAN, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.constants import TUE, WED, THU, FRI, SAT, SUN
from holidays.holiday_base import HolidayBase


class Chile(HolidayBase):
    """
    https://www.feriados.cl
    http://www.feriadoschilenos.cl/ (excellent history)
    https://es.wikipedia.org/wiki/Anexo:D%C3%ADas_feriados_en_Chile
    """

    country = "CL"
    special_holidays = {
        2022: ((SEP, 16, "Feriado nacional [National Holiday]"),),
    }
    # ISO 3166-2 codes for the principal subdivisions, called regions
    subdivisions = [
        "AI",
        "AN",
        "AP",
        "AR",
        "AT",
        "BI",
        "CO",
        "LI",
        "LL",
        "LR",
        "MA",
        "ML",
        "NB",
        "RM",
        "TA",
        "VS",
    ]

    def _populate(self, year):

        # Law 2.977 established official holidays for Chile
        # in its current form
        if year <= 1914:
            return
        super()._populate(year)

        # New Year's Day (Law 2.977)
        self[date(year, JAN, 1)] = "Año Nuevo [New Year's Day]"
        # Day after, if it's a Sunday (Law 20.983)
        if year >= 2017 and date(year, JAN, 1).weekday() == SUN:
            self[date(year, JAN, 2)] = "Fiestas Patrias [Holiday]"

        # Holy Week (Law 2.977)
        easter_date = easter(year)
        self[
            easter_date + rd(days=-2)
        ] = "Semana Santa (Viernes Santo) [Good Friday)]"
        self[
            easter_date + rd(days=-1)
        ] = "Semana Santa (Sábado Santo) [Good Saturday)]"

        # Ascension
        if year <= 1967:
            self[
                easter_date + rd(days=+39)
            ] = "Ascensión del Señor [Ascension of Jesus]"

        # Corpus Christi
        if year <= 1967 or 1987 <= year <= 2006:
            # Law 19.668
            if year <= 1999:
                dt = easter_date + rd(days=+60)
            else:
                dt = easter_date + rd(days=+57)
            self[dt] = "Corpus Christi [Corpus Christi]"

        # Labour Day (Law 2.200, renamed with Law 18.018)
        if year >= 1932:
            self[date(year, MAY, 1)] = "Día Nacional del Trabajo [Labour Day]"

        # Naval Glories Day (Law 2.977)
        self[date(year, MAY, 21)] = "Día de las Glorias Navales [Navy Day]"

        # National Day of Indigenous Peoples
        name = (
            "Día Nacional de los Pueblos Indígenas "
            "[National Day of Indigenous Peoples]"
        )
        if year == 2021:
            self[date(year, JUN, 21)] = name
        elif year >= 2022:
            epoch = Sun.get_equinox_solstice(year, target="summer")
            # Received date for UTC timezone needs to be adjusted
            # to match Chile's timezone
            # https://www.feriadoschilenos.cl/#DiaNacionalDeLosPueblosIndigenasII
            equinox = map(int, Epoch(epoch).get_full_date())
            adjusted_date = datetime(*equinox, tzinfo=tz.UTC).astimezone(
                tz.gettz("America/Santiago")
            )
            self[date(year, JUN, adjusted_date.day)] = name

        # Saint Peter and Saint Paul (Law 16.840, Law 18.432)
        if year <= 1967 or year >= 1986:
            dt = date(year, JUN, 29)
            if year >= 2000:
                # floating Monday holiday (Law 19.668)
                if dt.weekday() < FRI:
                    dt += rd(weekday=MO(-1))
                elif dt.weekday() == FRI:
                    dt += rd(weekday=MO)
            self[dt] = "San Pedro y San Pablo [Saint Peter and Saint Paul]"

        # Day of Virgin of Carmen (Law 20.148)
        if year >= 2007:
            self[
                date(year, JUL, 16)
            ] = "Virgen del Carmen [Our Lady of Mount Carmel]"

        # Day of Assumption of the Virgin (Law 2.977)
        self[
            date(year, AUG, 15)
        ] = "Asunción de la Virgen [Assumption of Mary]"

        # Day of National Liberation (Law 18.026)
        if 1981 <= year <= 1998:
            self[
                date(year, SEP, 11)
            ] = "Día de la Liberación Nacional [Day of National Liberation]"
        # Day of National Unity (Law 19.588, Law 19.793)
        elif 1999 <= year <= 2001:
            self[
                date(year, SEP, 1) + rd(weekday=MO)
            ] = "Día de la Unidad Nacional [Day of National Unity]"

        # National Holiday Friday preceding Independence Day (Law 20.983)
        # Monday, September 17, 2007, is declared a holiday.
        if year >= 2017 and date(year, SEP, 18).weekday() == SAT:
            self[date(year, SEP, 17)] = "Fiestas Patrias [Holiday]"

        # National Holiday Monday preceding Independence Day (Law 20.215)
        if year >= 2007 and date(year, SEP, 18).weekday() == TUE:
            self[date(year, SEP, 17)] = "Fiestas Patrias [Holiday]"

        # Independence Day (Law 2.977)
        self[
            date(year, SEP, 18)
        ] = "Día de la Independencia [Independence Day]"

        # Day of Glories of the Army of Chile (Law 2.977)
        self[
            date(year, SEP, 19)
        ] = "Día de las Glorias del Ejército [Army Day]"

        # National Holiday Friday following Army Day (Law 20.215)
        if year >= 2008 and date(year, SEP, 19).weekday() == THU:
            self[date(year, SEP, 20)] = "Fiestas Patrias [Holiday]"

        # Decree-law 636, Law 8.223
        if 1932 <= year <= 1944:
            self[date(year, SEP, 20)] = "Fiestas Patrias [Holiday]"

        # Columbus day (Law 3.810)
        if year >= 1922 and year != 1973:
            dt = date(year, OCT, 12)
            name = "Día de la Raza [Columbus day]"
            if year >= 2000:
                # floating Monday holiday (Law 19.668)
                if dt.weekday() < FRI:
                    dt += rd(weekday=MO(-1))
                elif dt.weekday() == FRI:
                    dt += rd(weekday=MO)
                if year <= 2019:
                    # Day of the Meeting of Two Worlds (Law 3.810)
                    name = (
                        "Día del Respeto a la Diversidad "
                        "[Day of the Meeting of Two Worlds]"
                    )
                else:
                    name = (
                        "Día del Descubrimiento de dos Mundos "
                        "[Discovery of Two Worlds' Day]"
                    )
            self[dt] = name

        # National Day of the Evangelical and Protestant Churches (Law 20.299)
        if year >= 2008:
            dt = date(year, OCT, 31)
            # This holiday is moved to the preceding Friday
            # if it falls on a Tuesday, or to the following Friday
            # if it falls on a Wednesday (Law 20.299)
            if dt.weekday() == WED:
                dt += rd(days=+2)
            elif dt.weekday() == TUE:
                dt += rd(days=-4)
            self[dt] = (
                "Día Nacional de las Iglesias Evangélicas y Protestantes"
                " [Reformation Day]"
            )

        # All Saints Day (Law 2.977)
        self[date(year, NOV, 1)] = "Día de Todos los Santos [All Saints Day]"

        # Immaculate Conception (Law 2.977)
        self[
            date(year, DEC, 8)
        ] = "La Inmaculada Concepción [Immaculate Conception]"

        if 1944 <= year <= 1988:
            self[date(year, DEC, 24)] = "Víspera de Navidad [Christmas Eve]"

        # Christmas (Law 2.977)
        self[date(year, DEC, 25)] = "Navidad [Christmas]"

        # región de Arica y Parinacota
        if self.subdiv == "AP" and year >= 2013:
            # Law 20.663
            self[date(year, JUN, 7)] = (
                "Asalto y Toma del Morro de Arica"
                " [Assault and Capture of Cape Arica]"
            )

        # región de Ñuble
        if self.subdiv == "NB" and year >= 2014:
            # Law 20.678
            self[date(year, AUG, 20)] = (
                "Nacimiento del Prócer de la Independencia"
                " (Chillán y Chillán Viejo)"
                " [Nativity of Bernardo O'Higgins]"
            )


class CL(Chile):
    pass


class CHL(Chile):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, APR, MAY, JUL, AUG, OCT
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class Turkey(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Turkey
    """

    country = "TR"

    def _populate(self, year):
        super()._populate(year)

        def _add_holiday(dt: date, hol: str) -> None:
            """Only add if in current year; prevents adding holidays across
            years (handles multi-day Islamic holidays that straddle Gregorian
            years).
            """
            if dt.year == year:
                self[dt] = hol

        # 1st of Jan
        self[date(year, JAN, 1)] = "New Year's Day"

        # 23rd of Apr
        self[date(year, APR, 23)] = "National Sovereignty and Children's Day"

        # 1st of May
        self[date(year, MAY, 1)] = "Labour Day"

        # 19th of May
        self[date(year, MAY, 19)] = (
            "Commemoration of Ataturk, Youth and " "Sports Day"
        )

        # 15th of Jul
        # Became a national holiday after 15 Jul 2016 coup d'etat attempt.
        if year > 2016:
            self[date(year, JUL, 15)] = "Democracy and National Unity Day"

        # 30th of Aug
        self[date(year, AUG, 30)] = "Victory Day"

        # 29th of Oct
        self[date(year, OCT, 29)] = "Republic Day"

        # Ramadan Feast
        # Date of observance is announced yearly, This is an estimate.
        for yr in (year - 1, year):
            for date_obs in _islamic_to_gre(yr, 10, 1):
                hol_date = date_obs
                _add_holiday(hol_date, "Ramadan Feast")
                _add_holiday(hol_date + rd(days=1), "Ramadan Feast Holiday")
                _add_holiday(hol_date + rd(days=2), "Ramadan Feast Holiday")

        # Sacrifice Feast
        # Date of observance is announced yearly, This is an estimate.
        for yr in (year - 1, year):
            for date_obs in _islamic_to_gre(yr, 12, 10):
                hol_date = date_obs
                _add_holiday(hol_date, "Sacrifice Feast")
                _add_holiday(hol_date + rd(days=1), "Sacrifice Feast Holiday")
                _add_holiday(hol_date + rd(days=2), "Sacrifice Feast Holiday")
                _add_holiday(hol_date + rd(days=3), "Sacrifice Feast Holiday")


class TR(Turkey):
    pass


class TUR(Turkey):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT
from holidays.constants import NOV, DEC
from holidays.holiday_base import HolidayBase


class Portugal(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Portugal
    """

    country = "PT"
    # https://en.wikipedia.org/wiki/Districts_of_Portugal
    # Only the 18 mainland districts have been included
    # `Ext` represents the national holidays most people have off
    subdivisions = [
        "Ext",
        "01",
        "02",
        "03",
        "04",
        "05",
        "06",
        "07",
        "08",
        "09",
        "10",
        "11",
        "12",
        "13",
        "14",
        "15",
        "16",
        "17",
        "18",
    ]

    def _populate(self, year):
        super()._populate(year)

        self[date(year, JAN, 1)] = "Ano Novo"

        easter_date = easter(year)

        # carnival is no longer a holiday, but some companies let workers off.
        # @todo recollect the years in which it was a public holiday
        # self[e - rd(days=47)] = "Carnaval"
        self[easter_date + rd(days=-2)] = "Sexta-feira Santa"
        self[easter_date] = "Páscoa"

        # Revoked holidays in 2013–2015
        if year < 2013 or year > 2015:
            self[easter_date + rd(days=+60)] = "Corpo de Deus"
            self[date(year, OCT, 5)] = "Implantação da República"
            self[date(year, NOV, 1)] = "Dia de Todos os Santos"
            self[date(year, DEC, 1)] = "Restauração da Independência"

        self[date(year, APR, 25)] = "Dia da Liberdade"
        self[date(year, MAY, 1)] = "Dia do Trabalhador"
        self[date(year, JUN, 10)] = "Dia de Portugal"
        self[date(year, AUG, 15)] = "Assunção de Nossa Senhora"
        self[date(year, DEC, 8)] = "Imaculada Conceição"
        self[date(year, DEC, 25)] = "Dia de Natal"

        if self.subdiv == "Ext":
            """
            Adds extended days that most people have as a bonus from their
            companies:

            - Carnival
            - the day before and after xmas
            - the day before the new year
            - Lisbon's city holiday
            """

            self[easter_date + rd(days=-47)] = "Carnaval"
            self[date(year, DEC, 24)] = "Véspera de Natal"
            self[date(year, DEC, 26)] = "26 de Dezembro"
            self[date(year, DEC, 31)] = "Véspera de Ano Novo"
            self[date(year, JUN, 13)] = "Dia de Santo António"

            # TODO add bridging days
            # - get Holidays that occur on Tuesday  and add Monday (-1 day)
            # - get Holidays that occur on Thursday and add Friday (+1 day)

        # District holidays
        if self.subdiv == "01":
            self[date(year, MAY, 12)] = "Dia de Santa Joana"
        if self.subdiv == "02":
            self[easter_date + rd(days=+39)] = "Quinta-feira da Ascensão"
        if self.subdiv in {"03", "13"}:
            self[date(year, JUN, 24)] = "Dia de São João"
        if self.subdiv == "04":
            self[date(year, AUG, 22)] = "Dia de Nossa Senhora das Graças"
        if self.subdiv == "05":
            self[
                easter_date + rd(days=+16)
            ] = "Dia de Nossa Senhora de Mércoles"
        if self.subdiv == "06":
            self[date(year, JUL, 4)] = "Dia de Santa Isabel"
        if self.subdiv == "07":
            self[date(year, JUN, 29)] = "Dia de S. Pedro"
        if self.subdiv == "08":
            self[date(year, SEP, 7)] = "Dia do Município de Faro"
        if self.subdiv == "09":
            self[date(year, NOV, 27)] = "Dia do Município da Guarda"
        if self.subdiv == "10":
            self[date(year, MAY, 22)] = "Dia do Município de Leiria"
        if self.subdiv in {"11", "17"}:
            self[date(year, JUN, 13)] = "Dia de Santo António"
        if self.subdiv == "12":
            self[date(year, MAY, 23)] = "Dia do Município de Portalegre"
        if self.subdiv == "14":
            self[date(year, MAR, 19)] = "Dia de S. José"
        if self.subdiv == "15":
            self[date(year, SEP, 15)] = "Dia de Bocage"
        if self.subdiv == "16":
            self[date(year, AUG, 20)] = "Dia de Nossa Senhora da Agonia"
        if self.subdiv == "18":
            self[date(year, SEP, 21)] = "Dia de S. Mateus"


class PT(Portugal):
    pass


class PRT(Portugal):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


from datetime import date

from convertdate import gregorian, hebrew
from convertdate.holidays import hanukkah, lag_baomer, passover, purim
from convertdate.holidays import rosh_hashanah, shavuot, sukkot, yom_kippur
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import WED, THU, SAT
from holidays.holiday_base import HolidayBase


class Israel(HolidayBase):

    country = "IL"

    def _populate(self, year):
        super()._populate(year)

        # Passover
        name = "Passover I"
        passover_start_dt = date(*passover(year, eve=True))
        self[passover_start_dt] = name + " - Eve"
        self[passover_start_dt + rd(days=1)] = name

        name = "Passover"
        for offset in range(2, 6):
            self[passover_start_dt + rd(days=offset)] = name + " - Chol HaMoed"

        name = "Passover VII"
        self[passover_start_dt + rd(days=6)] = name + " - Eve"
        self[passover_start_dt + rd(days=7)] = name

        # Memorial Day
        name = "Memorial Day"
        memorial_day_dt = date(
            *gregorian.from_jd(
                hebrew.to_jd_gregorianyear(year, hebrew.IYYAR, 3)
            )
        )
        self[memorial_day_dt + rd(days=1)] = name

        observed_delta = 0
        if self.observed:
            day_in_week = memorial_day_dt.weekday()
            if day_in_week in {WED, THU}:
                observed_delta = -(day_in_week - 1)
            elif 2004 <= year and day_in_week == SAT:
                observed_delta = 1

            if observed_delta != 0:
                self[memorial_day_dt + rd(days=observed_delta + 1)] = (
                    name + " (Observed)"
                )

        # Independence Day
        name = "Independence Day"
        self[memorial_day_dt + rd(days=2)] = name

        if self.observed and observed_delta != 0:
            self[memorial_day_dt + rd(days=observed_delta + 2)] = (
                name + " (Observed)"
            )

        # Lag Baomer
        name = "Lag B'Omer"
        lag_baomer_dt = date(*lag_baomer(year, eve=False))
        self[lag_baomer_dt] = name

        # Shavuot
        name = "Shavuot"
        shavuot_dt = date(*shavuot(year, eve=True))
        self[shavuot_dt] = name + " - Eve"
        self[shavuot_dt + rd(days=1)] = name

        # Rosh Hashana
        name = "Rosh Hashanah"
        rosh_hashanah_dt = date(*rosh_hashanah(year, eve=True))
        self[rosh_hashanah_dt] = name + " - Eve"
        self[rosh_hashanah_dt + rd(days=1)] = name
        self[rosh_hashanah_dt + rd(days=2)] = name

        # Yom Kippur
        name = "Yom Kippur"
        yom_kippur_dt = date(*yom_kippur(year, eve=True))
        self[yom_kippur_dt] = name + " - Eve"
        self[yom_kippur_dt + rd(days=1)] = name

        # Sukkot
        name = "Sukkot I"
        sukkot_start_dt = date(*sukkot(year, eve=True))
        self[sukkot_start_dt] = name + " - Eve"
        self[sukkot_start_dt + rd(days=1)] = name

        name = "Sukkot"
        for offset in range(2, 7):
            self[sukkot_start_dt + rd(days=offset)] = name + " - Chol HaMoed"

        name = "Sukkot VII"
        self[sukkot_start_dt + rd(days=7)] = name + " - Eve"
        self[sukkot_start_dt + rd(days=8)] = name

        # Hanukkah
        name = "Hanukkah"
        hk_start_date = date(*hanukkah(year, eve=False))
        for offset in range(8):
            hk_date = hk_start_date + rd(days=offset)
            if hk_date.year == year:
                self[hk_date] = name
        # Some o prior's year Hannukah may fall in current year.
        hk_start_date = date(*hanukkah(year - 1, eve=False))
        for offset in range(8):
            hk_date = hk_start_date + rd(days=offset)
            if hk_date.year == year:
                self[hk_date] = name

        # Purim
        name = "Purim"
        purim_date = date(*purim(year, eve=True))
        self[purim_date] = name + " - Eve"
        self[purim_date + rd(days=1)] = name
        self[purim_date + rd(days=2)] = "Shushan Purim"


class IL(Israel):
    pass


class ISR(Israel):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, AUG, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase


class Austria(HolidayBase):
    country = "AT"
    subdivisions = ["1", "2", "3", "4", "5", "6", "7", "8", "9"]

    def __init__(self, **kwargs):
        if isinstance(kwargs.get("subdiv"), int):
            kwargs["subdiv"] = str(kwargs["subdiv"])
        # Default subdivision to "9"
        if not kwargs.get("subdiv", kwargs.get("state")):
            kwargs["subdiv"] = "9"
        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # public holidays
        self[date(year, JAN, 1)] = "Neujahr"
        self[date(year, JAN, 6)] = "Heilige Drei Könige"
        easter_date = easter(year)
        self[easter_date + rd(days=+1)] = "Ostermontag"
        self[date(year, MAY, 1)] = "Staatsfeiertag"
        self[easter_date + rd(days=+39)] = "Christi Himmelfahrt"
        self[easter_date + rd(days=+50)] = "Pfingstmontag"
        self[easter_date + rd(days=+60)] = "Fronleichnam"
        self[date(year, AUG, 15)] = "Mariä Himmelfahrt"
        if 1919 <= year <= 1934:
            self[date(year, NOV, 12)] = "Nationalfeiertag"
        if year >= 1967:
            self[date(year, OCT, 26)] = "Nationalfeiertag"
        self[date(year, NOV, 1)] = "Allerheiligen"
        self[date(year, DEC, 8)] = "Mariä Empfängnis"
        self[date(year, DEC, 25)] = "Christtag"
        self[date(year, DEC, 26)] = "Stefanitag"


class AT(Austria):
    pass


class AUT(Austria):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.relativedelta import relativedelta as rd

# Installation: pip install korean_lunar_calendar
# URL: https://github.com/usingsky/korean_lunar_calendar_py/
from korean_lunar_calendar import KoreanLunarCalendar

from holidays.constants import JAN, APR, MAY, SEP
from holidays.holiday_base import HolidayBase


class Vietnam(HolidayBase):
    """
    https://publicholidays.vn/
    http://vbpl.vn/TW/Pages/vbpqen-toanvan.aspx?ItemID=11013 Article.115
    https://www.timeanddate.com/holidays/vietnam/
    """

    country = "VN"

    def __init__(self, **kwargs):
        self.korean_cal = KoreanLunarCalendar()
        HolidayBase.__init__(self, **kwargs)

    def _add_observed(self, holiday: date) -> None:
        if self._is_weekend(holiday):
            next_workday = holiday + rd(days=+1)
            while self._is_weekend(next_workday) or self.get(next_workday):
                next_workday += rd(days=+1)
            self[next_workday] = self[holiday] + " observed"

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self[date(year, JAN, 1)] = "International New Year's Day"

        # Vietnamese Kings' Commemoration Day
        # https://en.wikipedia.org/wiki/H%C3%B9ng_Kings%27_Festival
        if year >= 2007:
            hol_date = self.get_solar_date(year, 3, 10)
            self[hol_date] = "Hung Kings Commemoration Day"

        # Liberation Day/Reunification Day
        self[date(year, APR, 30)] = "Liberation Day/Reunification Day"

        # International Labor Day
        self[date(year, MAY, 1)] = "International Labor Day"

        # Independence Day
        self[date(year, SEP, 2)] = "Independence Day"

        if self.observed:
            for dt in sorted(list(self.keys())):
                if dt.year == year:
                    self._add_observed(dt)

        # Lunar New Year
        names = (
            (-1, "Vietnamese New Year's Eve"),
            (0, "Vietnamese New Year"),
            (1, "The second day of Tet Holiday"),
            (2, "The third day of Tet Holiday"),
            (3, "The forth day of Tet Holiday"),
            (4, "The fifth day of Tet Holiday"),
        )
        hol_date = self.get_solar_date(year, 1, 1)
        for d, name in names:
            self[(hol_date + rd(days=+d))] = name

    # convert lunar calendar date to solar
    def get_solar_date(self, year, month, day):
        self.korean_cal.setLunarDate(year, month, day, False)
        return date(
            self.korean_cal.solarYear,
            self.korean_cal.solarMonth,
            self.korean_cal.solarDay,
        )


class VN(Vietnam):
    pass


class VNM(Vietnam):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import OCT, DEC
from holidays.holiday_base import HolidayBase


class Paraguay(HolidayBase):
    """
    https://www.ghp.com.py/news/feriados-nacionales-del-ano-2019-en-paraguay
    https://es.wikipedia.org/wiki/Anexo:D%C3%ADas_feriados_en_Paraguay
    http://www.calendarioparaguay.com/
    """

    country = "PY"
    special_holidays = {
        # public holiday for business purposes, in view of
        # the recently increased risk of Dengue fever
        2007: ((JAN, 29, "Public Holiday"),),
        # public sector holiday to celebrate Paraguay's
        # football team's qualification for the 2010 World Cup
        2009: ((SEP, 10, "Public Holiday"),),
        2010: (
            # public holiday to coincide with the Paraguay-Italy
            # game of the current World Cup
            (JUN, 14, "Public Holiday"),
            # 2 year-end public sector holidays
            (DEC, 24, "Public sector Holiday"),
            (DEC, 31, "Public sector Holiday"),
        ),
        2011: (
            # public holiday to coincide with the current anti-Dengue drive
            (APR, 19, "Public Holiday"),
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 20, "Public sector Holiday"),
            # public holidays to commemorate the Bicentennial
            # of Paraguay's independence
            (MAY, 14, "Public Holiday"),
            (MAY, 16, "Public Holiday"),
            # 2 year-end public sector holidays
            (DEC, 23, "Public sector Holiday"),
            (DEC, 30, "Public sector Holiday"),
        ),
        2012: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 4, "Public sector Holiday"),
            # 2 year-end public sector holidays
            (DEC, 24, "Public sector Holiday"),
            (DEC, 31, "Public sector Holiday"),
        ),
        2013: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (MAR, 27, "Public sector Holiday"),
            # date of the inauguration of President-elect
            # Horacio Cartes, as a one-off non-working public holiday
            (AUG, 14, "Public Holiday"),
        ),
        2014: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 16, "Public sector Holiday"),
            # 2 year-end public sector holidays
            (DEC, 24, "Public sector Holiday"),
            (DEC, 31, "Public sector Holiday"),
        ),
        2015: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 1, "Public sector Holiday"),
            # public holidays in Paraguay on account
            # of the upcoming visit of Pope Francis in Paraguay
            (JUL, 10, "Public Holiday"),
            (JUL, 11, "Public Holiday"),
            # 2 year-end public sector holidays
            (DEC, 24, "Public sector Holiday"),
            (DEC, 31, "Public sector Holiday"),
        ),
        2016: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (MAR, 23, "Public sector Holiday"),
            (DEC, 26, "Public Holiday"),
        ),
        2017: (
            (JAN, 2, "Public Holiday"),
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (MAR, 28, "Public sector Holiday"),
        ),
        2018: (
            # 2 year-end public sector holidays
            (DEC, 24, "Public sector Holiday"),
            (DEC, 31, "Public sector Holiday"),
        ),
        2019: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 17, "Public sector Holiday"),
            # 2 year-end public sector holidays
            (DEC, 24, "Public sector Holiday"),
            (DEC, 31, "Public sector Holiday"),
        ),
        2020: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 8, "Public sector Holiday"),
        ),
        2021: (
            # 2 year-end public sector holidays
            (DEC, 24, "Public sector Holiday"),
            (DEC, 31, "Public sector Holiday"),
        ),
        2022: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 13, "Public sector Holiday"),
            # public sector holiday due to the annual May 1st
            # public holiday falling on a Sunday
            (MAY, 2, "Public Holiday"),
        ),
    }

    def _add_holiday(self, dt: date, name: str) -> None:
        if self.observed or not self._is_weekend(dt):
            self[dt] = name

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self._add_holiday(date(year, JAN, 1), "Año Nuevo [New Year's Day]")

        # Patriots day
        dates_obs = {
            2013: (MAR, 4),
            2016: (FEB, 29),
            2018: (FEB, 26),
            2022: (FEB, 28),
        }
        self[
            date(year, *dates_obs.get(year, (MAR, 1)))
        ] = "Día de los Héroes de la Patria [Patriots Day]"

        # Holy Week
        easter_date = easter(year)
        self[easter_date + rd(days=-3)] = "Jueves Santo [Maundy Thursday]"
        self[easter_date + rd(days=-2)] = "Viernes Santo [Good Friday]"
        self._add_holiday(easter_date, "Día de Pascuas [Easter Day]")

        # Labor Day
        self._add_holiday(
            date(year, MAY, 1), "Día del Trabajador [Labour Day]"
        )

        # Independence Day
        name = "Día de la Independencia Nacional [Independence Day]"
        if year == 2021:
            self[date(year, MAY, 14)] = name
            self[date(year, MAY, 15)] = name
        elif year >= 2012:
            self._add_holiday(date(year, MAY, 14), name)
            self._add_holiday(date(year, MAY, 15), name)
        else:
            self._add_holiday(date(year, MAY, 15), name)

        # Peace in Chaco Day.
        dates_obs = {
            2014: (JUN, 16),
            2018: (JUN, 11),
        }
        self._add_holiday(
            date(year, *dates_obs.get(year, (JUN, 12))),
            "Día de la Paz del Chaco [Chaco Armistice Day]",
        )

        # Asuncion Fundation's Day
        self._add_holiday(
            date(year, AUG, 15),
            "Día de la Fundación de Asunción [Asuncion Foundation's Day]",
        )

        # Boqueron's Battle
        if year >= 2000:
            dates_obs = {
                2015: (SEP, 28),
                2016: (OCT, 3),
                2017: (OCT, 2),
                2021: (SEP, 27),
                2022: (OCT, 3),
            }
            self._add_holiday(
                date(year, *dates_obs.get(year, (SEP, 29))),
                "Día de la Batalla de Boquerón [Boqueron Battle Day]",
            )

        # Caacupe Virgin Day
        self._add_holiday(
            date(year, DEC, 8),
            "Día de la Virgen de Caacupé [Caacupe Virgin Day]",
        )

        # Christmas
        self[date(year, DEC, 25)] = "Navidad [Christmas]"


class PY(Paraguay):
    pass


class PRY(Paraguay):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from typing import Iterable, Optional, Union

from dateutil.easter import easter
from dateutil.relativedelta import FR, MO, SA, SU
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import OCT, NOV, DEC, FRI, SAT, SUN
from holidays.holiday_base import HolidayBase
from holidays.utils import _ChineseLuniSolar, _islamic_to_gre


class Malaysia(HolidayBase):
    country = "MY"
    special_holidays = {
        # The years 1955 1959 1995 seems to have the elections
        # one weekday but I am not sure if they were marked as
        # holidays.
        1999: ((NOV, 29, "Malaysia General Election Holiday"),),
        2018: ((MAY, 9, "Malaysia General Election Holiday"),),
        2019: ((JUL, 30, "Installation of New King"),),
    }
    subdivisions = [
        "JHR",
        "KDH",
        "KTN",
        "MLK",
        "NSN",
        "PHG",
        "PRK",
        "PLS",
        "PNG",
        "SBH",
        "SWK",
        "SGR",
        "TRG",
        "KUL",
        "LBN",
        "PJY",
    ]

    def __init__(
        self,
        years: Optional[Union[int, Iterable[int]]] = None,
        expand: bool = True,
        observed: bool = True,
        subdiv: Optional[str] = None,
        prov: Optional[str] = None,
        state: Optional[str] = None,
    ) -> None:
        """
        An subclass of :py:class:`HolidayBase` representing public holidays in
        Malaysia.

        If ``subdiv`` for a state is not supplied, only nationwide holidays are
        returned. The following ``subdiv`` state codes are used (ISO 3166-2
        subdivision codes are not yet supported):

        - JHR: Johor
        - KDH: Kedah
        - KTN: Kelantan
        - MLK: Melaka
        - NSN: Negeri Sembilan
        - PHG: Pahang
        - PRK: Perak
        - PLS: Perlis
        - PNG: Pulau Pinang
        - SBH: Sabah
        - SWK: Sarawak
        - SGR: Selangor
        - TRG: Terengganu
        - KUL: FT Kuala Lumpur
        - LBN: FT Labuan
        - PJY: FT Putrajaya

        Limitations:

        - Prior to 2021: holidays are not accurate.
        - 2027 and later: Thaipusam dates are are estimated, and so denoted.

        Reference: `Wikipedia
        <https://en.wikipedia.org/wiki/Public_holidays_in_Malaysia>`__

        Country created by: `Eden <https://github.com/jusce17>`__

        Country maintained by: `Mike Borsetti <https://github.com/mborsetti>`__

        See parameters and usage in :py:class:`HolidayBase`.
        """
        self.cnls = _ChineseLuniSolar()
        super().__init__(years, expand, observed, subdiv, prov, state)

    def _populate(self, year):
        def _add_holiday(dt: date, hol: str) -> None:
            if dt.year == year:
                self[dt] = hol

        super()._populate(year)

        estimated_suffix = "* (*estimated)"

        # Chinese New Year (one day in the States of Kelantan and Terengganu,
        # two days in the other States).
        hol_date = self.cnls.lunar_n_y_date(year)
        self[hol_date] = "Chinese New Year"
        # The second day of Chinese New Year is not a federal holiday in
        # Kelantan and Terengganu. However, it is gazetted as a state holiday
        # in both states, effectively making it a nationwide holiday.
        self[hol_date + rd(days=+1)] = "Chinese New Year Holiday"

        # Wesak Day.
        # Date of observance is announced yearly
        # https://en.wikipedia.org/wiki/Vesak#Dates_of_observance
        dates_obs = {
            2001: (MAY, 7),
            2002: (MAY, 27),
            2003: (MAY, 15),
            2004: (MAY, 3),
            2005: (MAY, 23),
            2006: (MAY, 12),
            2007: (MAY, 1),
            2008: (MAY, 19),
            2009: (MAY, 9),
            2010: (MAY, 28),
            2011: (MAY, 17),
            2012: (MAY, 5),
            2013: (MAY, 24),
            2014: (MAY, 13),
            2015: (MAY, 3),
            2016: (MAY, 21),
            2017: (MAY, 10),
            2018: (MAY, 29),
            2019: (MAY, 19),
            2020: (MAY, 7),
            2021: (MAY, 26),
            2022: (MAY, 15),
        }
        name = "Vesak Day"
        if year in dates_obs:
            hol_date = date(year, *dates_obs[year])
        else:
            hol_date = self.cnls.vesak_may_date(year)
            name += estimated_suffix
        self[hol_date] = name

        # Labour Day.
        self[date(year, MAY, 1)] = "Labour Day"

        # Birthday of [His Majesty] the Yang di-Pertuan Agong.
        if year <= 2017:
            hol_date = date(year, JUN, 1) + rd(weekday=SA)
        elif year == 2018:
            hol_date = date(2018, SEP, 9)
        elif year == 2020:
            # https://www.nst.com.my/news/nation/2020/03/571660/agongs-birthday-moved-june-6-june-8
            hol_date = date(2020, JUN, 8)
        else:
            hol_date = date(year, JUN, 1) + rd(weekday=MO)
        self[hol_date] = "Birthday of SPB Yang di-Pertuan Agong"

        # Hari Kebangsaan or National Day.
        self[date(year, AUG, 31)] = "National Day"

        # Malaysia Day.
        if year >= 2010:
            self[date(year, SEP, 16)] = "Malaysia Day"

        # Deepavali.
        # aka Diwali;
        # date of observance is announced yearly
        if self.subdiv != "SWK":
            dates_obs = {
                2001: (NOV, 14),
                2002: (NOV, 3),
                2003: (OCT, 23),
                2004: (NOV, 11),
                2005: (NOV, 1),
                2006: (OCT, 21),
                2007: (NOV, 8),
                2008: (OCT, 27),
                2009: (OCT, 17),
                2010: (NOV, 5),
                2011: (OCT, 26),
                2012: (NOV, 13),
                2013: (NOV, 2),
                2014: (OCT, 22),
                2015: (NOV, 10),
                2016: (OCT, 29),
                2017: (OCT, 18),
                2018: (NOV, 6),
                2019: (OCT, 27),
                2020: (NOV, 14),
                2021: (NOV, 4),
                2022: (OCT, 24),
            }
            name = "Deepavali"
            if year in dates_obs:
                hol_date = date(year, *dates_obs[year])
            else:
                hol_date = self.cnls.s_diwali_date(year)
                name += estimated_suffix
            self[hol_date] = name

        # Christmas day.
        self[date(year, DEC, 25)] = "Christmas Day"

        # Birthday of the Prophet Muhammad (s.a.w.).
        # a.k.a. Hari Keputeraan Nabi Muhammad (Sabah Act)
        dates_obs = {
            2001: ((JUN, 4),),
            2002: ((MAY, 24),),
            2003: ((MAY, 14),),
            2004: ((MAY, 2),),
            2005: ((APR, 21),),
            2006: ((APR, 11),),
            2007: ((MAR, 31),),
            2008: ((MAR, 20),),
            2009: ((MAR, 9),),
            2010: ((FEB, 26),),
            2011: ((FEB, 16),),
            2012: ((FEB, 5),),
            2013: ((JAN, 24),),
            2014: ((JAN, 14),),
            2015: (
                (JAN, 3),
                (DEC, 24),
            ),
            2016: ((DEC, 12),),
            2017: ((DEC, 1),),
            2018: ((NOV, 20),),
            2019: ((NOV, 9),),
            2020: ((OCT, 29),),
            2021: ((OCT, 19),),
            2022: ((OCT, 10),),
        }
        name = "Maulidur Rasul (Birthday of the Prophet Muhammad)"
        if year in dates_obs:
            hol_dates = [
                (date(year, *date_obs), "") for date_obs in dates_obs[year]
            ]
        else:
            hol_dates = [
                (date_obs, estimated_suffix)
                for date_obs in _islamic_to_gre(year, 3, 12)
            ]
        for hol_date, hol_suffix in hol_dates:
            self[hol_date] = name + hol_suffix

        # Hari Raya Puasa (2 days).
        # aka Eid al-Fitr;
        # exact date of observance is announced yearly
        dates_obs = {
            2001: ((DEC, 17),),
            2002: ((DEC, 6),),
            2003: ((NOV, 26),),
            2004: ((NOV, 14),),
            2005: ((NOV, 3),),
            2006: ((OCT, 24),),
            2007: ((OCT, 13),),
            2008: ((OCT, 1),),
            2009: ((SEP, 20),),
            2010: ((SEP, 10),),
            2011: ((AUG, 31),),
            2012: ((AUG, 19),),
            2013: ((AUG, 8),),
            2014: ((JUL, 28),),
            2015: ((JUL, 17),),
            2016: ((JUL, 6),),
            2017: ((JUN, 25),),
            2018: ((JUN, 15),),
            2019: ((JUN, 5),),
            2020: ((MAY, 24),),
            2021: ((MAY, 13),),
            2022: ((MAY, 2),),
        }
        name = "Hari Raya Puasa"
        hol_dates = []
        for yr in (year - 1, year):
            if yr in dates_obs:
                for date_obs in dates_obs[yr]:
                    hol_dates.append((date(yr, *date_obs), ""))
            else:
                for date_obs in _islamic_to_gre(year, 10, 1):
                    hol_dates.append((date_obs, estimated_suffix))
        for hol_date, hol_suffix in hol_dates:
            _add_holiday(hol_date, f"{name}{hol_suffix}")
            _add_holiday(
                hol_date + rd(days=+1), f"Second day of {name}{hol_suffix}"
            )

        # Hari Raya Haji and Arafat Day.
        # Date of observance is announced yearly.
        dates_obs = {
            2001: ((MAR, 6),),
            2002: ((FEB, 23),),
            2003: ((FEB, 12),),
            2004: ((FEB, 2),),
            2005: ((JAN, 21),),
            2006: (
                (JAN, 10),
                (DEC, 31),
            ),
            2007: ((DEC, 20),),
            2008: ((DEC, 9),),
            2009: ((NOV, 28),),
            2010: ((NOV, 17),),
            2011: ((NOV, 7),),
            2012: ((OCT, 26),),
            2013: ((OCT, 15),),
            2014: ((OCT, 5),),
            2015: ((SEP, 24),),
            2016: ((SEP, 12),),
            2017: ((SEP, 1),),
            2018: ((AUG, 22),),
            2019: ((AUG, 11),),
            2020: ((JUL, 31),),
            2021: ((JUL, 20),),
            2022: ((JUL, 10),),
        }
        name = "Hari Raya Haji"
        hol_dates = []
        for yr in (year - 1, year):
            if yr in dates_obs:
                for date_obs in dates_obs[yr]:
                    hol_dates.append((date(yr, *date_obs), ""))
            else:
                for date_obs in _islamic_to_gre(year, 12, 10):
                    hol_dates.append((date_obs, estimated_suffix))
        for hol_date, hol_suffix in hol_dates:
            _add_holiday(hol_date, f"{name}{hol_suffix}")
            if self.subdiv == "TRG":
                # Arafat Day is one day before Eid al-Adha
                _add_holiday(hol_date + rd(days=-1), f"Arafat Day{hol_suffix}")
            if self.subdiv in {"KDH", "KTN", "PLS", "TRG"}:
                # Second day
                _add_holiday(
                    hol_date + rd(days=+1), f"{name} Holiday{hol_suffix}"
                )

        # ---------------------------------------------------------#
        # Holidays from the Sarawak Ordinance (not included above) #
        # ---------------------------------------------------------#
        if self.subdiv == "SWK":
            # Dayak Festival Day (the first day of June) and the following day.
            self[date(year, JUN, 1)] = "Gawai Dayak"
            self[date(year, JUN, 2)] = "Gawai Dayak (Second day)"

            # The first day of May—Worker’s Celebration Day.

            # Birthday of Tuan Yang Terutama Yang di-Pertua Negeri Sarawak (the
            # second Saturday of September).
            second_sat_oct = date(year, OCT, 1) + rd(weekday=SA(+2))
            self[second_sat_oct] = "Birthday of the Governor of Sarawak"

            # Sarawak Independence Day
            if year >= 2017:
                self[date(year, JUL, 22)] = "Sarawak Day"

        # New Year's Day
        if self.subdiv in {
            "KUL",
            "LBN",
            "MLK",
            "NSN",
            "PHG",
            "PJY",
            "PNG",
            "PRK",
            "SBH",
            "SGR",
            "SWK",
        }:
            self[date(year, JAN, 1)] = "New Year's Day"

        # Isra and Mi'raj.
        if self.subdiv in {"KDH", "NSN", "PLS"} or (
            self.subdiv == "TRG" and year >= 2020
        ):
            dates_obs = {
                2001: ((OCT, 15),),
                2002: ((OCT, 4),),
                2003: ((SEP, 24),),
                2004: ((SEP, 12),),
                2005: ((SEP, 1),),
                2006: ((AUG, 22),),
                2007: ((AUG, 11),),
                2008: ((JUL, 31),),
                2009: ((JUL, 20),),
                2010: ((JUL, 9),),
                2011: ((JUN, 29),),
                2012: ((JUN, 17),),
                2013: ((JUN, 6),),
                2014: ((MAY, 27),),
                2015: ((MAY, 16),),
                2016: ((MAY, 5),),
                2017: ((APR, 24),),
                2018: ((APR, 14),),
                2019: ((APR, 3),),
                2020: ((MAR, 22),),
                2021: ((MAR, 11),),
                2022: ((MAR, 1),),
            }
            name = "Isra and Mi'raj"
            if year in dates_obs:
                hol_dates = [
                    (date(year, *date_obs), "") for date_obs in dates_obs[year]
                ]
            else:
                hol_dates = [
                    (date_obs, estimated_suffix)
                    for date_obs in _islamic_to_gre(year, 7, 27)
                ]
            for hol_date, hol_suffix in hol_dates:
                self[hol_date] = name + hol_suffix

        # Beginning of Ramadan.
        if self.subdiv in {"JHR", "KDH", "MLK"}:
            dates_obs = {
                2001: ((NOV, 17),),
                2002: ((NOV, 6),),
                2003: ((OCT, 27),),
                2004: ((OCT, 16),),
                2005: ((OCT, 5),),
                2006: ((SEP, 24),),
                2007: ((SEP, 13),),
                2008: ((SEP, 2),),
                2009: ((AUG, 22),),
                2010: ((AUG, 11),),
                2011: ((AUG, 1),),
                2012: ((JUL, 20),),
                2013: ((JUL, 9),),
                2014: ((JUN, 29),),
                2015: ((JUN, 18),),
                2016: ((JUN, 7),),
                2017: ((MAY, 27),),
                2018: ((MAY, 17),),
                2019: ((MAY, 6),),
                2020: ((APR, 24),),
                2021: ((APR, 13),),
                2022: ((APR, 3),),
            }
            name = "Beginning of Ramadan"
            if year in dates_obs:
                hol_dates = [
                    (date(year, *date_obs), "") for date_obs in dates_obs[year]
                ]
            else:
                hol_dates = [
                    (date_obs, estimated_suffix)
                    for date_obs in _islamic_to_gre(year, 9, 1)
                ]
            for hol_date, hol_suffix in hol_dates:
                self[hol_date] = name + hol_suffix

        # Nuzul Al-Quran Day.
        if self.subdiv in {
            "KTN",
            "KUL",
            "LBN",
            "PHG",
            "PNG",
            "PRK",
            "PLS",
            "PJY",
            "SGR",
            "TRG",
        }:
            dates_obs = {
                2001: ((DEC, 3),),
                2002: ((NOV, 22),),
                2003: ((NOV, 12),),
                2004: ((NOV, 1),),
                2005: ((OCT, 21),),
                2006: ((OCT, 10),),
                2007: ((SEP, 29),),
                2008: ((SEP, 18),),
                2009: ((SEP, 7),),
                2010: ((AUG, 27),),
                2011: ((AUG, 17),),
                2012: ((AUG, 5),),
                2013: ((JUL, 25),),
                2014: ((JUL, 15),),
                2015: ((JUL, 4),),
                2016: ((JUN, 22),),
                2017: ((JUN, 12),),
                2018: ((JUN, 2),),
                2019: ((MAY, 22),),
                2020: ((MAY, 10),),
                2021: ((APR, 29),),
                2022: ((APR, 19),),
            }
            name = "Nuzul Al-Quran Day"
            if year in dates_obs:
                hol_dates = [
                    (date(year, *date_obs), "") for date_obs in dates_obs[year]
                ]
            else:
                hol_dates = [
                    (date_obs, estimated_suffix)
                    for date_obs in _islamic_to_gre(year, 9, 17)
                ]
            for hol_date, hol_suffix in hol_dates:
                self[hol_date] = name + hol_suffix

        # Thaipusam.
        # An annual Hindu festival observed on the day of the first full moon
        # during the Tamil month of Thai
        if self.subdiv in {"JHR", "KUL", "NSN", "PJY", "PNG", "PRK", "SGR"}:
            dates_obs = {
                2018: (JAN, 31),
                2019: (JAN, 21),
                2020: (FEB, 8),
                2021: (JAN, 28),
                2022: (JAN, 18),
                2023: (FEB, 5),
                2024: (JAN, 25),
                2025: (FEB, 11),
                2026: (FEB, 1),
                2027: (JAN, 22),
            }
            name = "Thaipusam"
            if year in dates_obs:
                hol_date = date(year, *dates_obs[year])
            else:
                hol_date = self.cnls.thaipusam_date(year)
                name += estimated_suffix
            self[hol_date] = name

        # Federal Territory Day.
        if self.subdiv in {"KUL", "LBN", "PJY"} and year >= 1974:
            self[date(year, FEB, 1)] = "Federal Territory Day"

        # State holidays (single state)

        if self.subdiv == "MLK":
            if year >= 1989:
                self[
                    date(year, APR, 15)
                ] = "Declaration of Malacca as a Historical City"
            if year >= 2020:
                hol_date = date(year, AUG, 24)
            else:
                hol_date = date(year, OCT, 1) + rd(weekday=FR(+2))
            self[hol_date] = "Birthday of the Governor of Malacca"

        elif self.subdiv == "NSN":
            if year >= 2009:
                self[
                    date(year, JAN, 14)
                ] = "Birthday of the Sultan of Negeri Sembilan"

        elif self.subdiv == "PHG":
            if year >= 2020:
                hol_date = date(year, MAY, 22)
            else:
                hol_date = date(year, MAY, 7)
            self[hol_date] = "Hari Hol of Pahang"

            if year >= 2019:
                hol_date = date(year, JUL, 30)
            elif 1975 <= year <= 2018:
                hol_date = date(year, OCT, 24)
            self[hol_date] = "Birthday of the Sultan of Pahang"

        elif self.subdiv == "PNG":
            if year >= 2009:
                self[date(year, JUL, 7)] = "George Town Heritage Day"
            second_sat_jul = date(year, JUL, 1) + rd(weekday=SA(+2))
            self[second_sat_jul] = "Birthday of the Governor of Penang"

        elif self.subdiv == "PLS":
            if year >= 2018:
                hol_date = date(year, JUL, 17)
            elif year >= 2000:
                hol_date = date(year, MAY, 17)
            self[hol_date] = "Birthday of The Raja of Perlis"

        elif self.subdiv == "SGR":
            self[date(year, DEC, 11)] = "Birthday of The Sultan of Selangor"

        elif self.subdiv == "TRG":
            if year >= 2000:
                self[date(year, MAR, 4)] = (
                    "Anniversary of the Installation "
                    "of the Sultan of Terengganu"
                )
                self[
                    date(year, APR, 26)
                ] = "Birthday of the Sultan of Terengganu"

        # Check for holidays that fall on a Sunday and
        # implement Section 3 of Malaysian Holidays Act:
        # "if any day specified in the Schedule falls on
        # Sunday then the day following shall be a public
        # holiday and if such day is already a public holiday,
        # then the day following shall be a public holiday"
        # In Johor and Kedah it's Friday -> Sunday,
        # in Kelantan and Terengganu it's Saturday -> Sunday
        for hol_date, hol_name in list(self.items()):
            if hol_date.year != year:
                continue
            in_lieu_date = None
            if hol_date.weekday() == FRI and self.subdiv in {"JHR", "KDH"}:
                in_lieu_date = hol_date + rd(days=+2)
            elif hol_date.weekday() == SAT and self.subdiv in {
                "KTN",
                "TRG",
            }:
                in_lieu_date = hol_date + rd(days=+1)
            elif hol_date.weekday() == SUN and self.subdiv not in {
                "JHR",
                "KDH",
                "KTN",
                "TRG",
            }:
                in_lieu_date = hol_date + rd(days=+1)
            if not in_lieu_date:
                continue
            while in_lieu_date.year == year and in_lieu_date in self:
                in_lieu_date += rd(days=+1)
            _add_holiday(in_lieu_date, f"{hol_name} [In lieu]")

        # The last two days in May (Pesta Kaamatan).
        # (Sarawak Act)
        # Day following a Sunday is not a holiday
        if self.subdiv in {"LBN", "SBH"}:
            self[date(year, MAY, 30)] = "Pesta Kaamatan"
            self[date(year, MAY, 31)] = "Pesta Kaamatan (Second day)"

        # ------------------------------#
        # Other holidays (decrees etc.) #
        # ------------------------------#

        # Awal Muharram.
        dates_obs = {
            2001: ((MAR, 26),),
            2002: ((MAR, 15),),
            2003: ((MAR, 5),),
            2004: ((FEB, 22),),
            2005: ((FEB, 10),),
            2006: ((JAN, 31),),
            2007: ((JAN, 20),),
            2008: (
                (JAN, 10),
                (DEC, 29),
            ),
            2009: ((DEC, 18),),
            2010: ((DEC, 8),),
            2011: ((NOV, 27),),
            2012: ((NOV, 15),),
            2013: ((NOV, 5),),
            2014: ((OCT, 25),),
            2015: ((OCT, 14),),
            2016: ((OCT, 2),),
            2017: ((SEP, 22),),
            2018: ((SEP, 11),),
            2019: ((SEP, 1),),
            2020: ((AUG, 20),),
            2021: ((AUG, 10),),
            2022: ((JUL, 30),),
        }
        name = "Awal Muharram (Hijri New Year)"
        if year in dates_obs:
            hol_dates = [
                (date(year, *date_obs), "") for date_obs in dates_obs[year]
            ]
        else:
            hol_dates = [
                (date_obs, estimated_suffix)
                for date_obs in _islamic_to_gre(year, 1, 1)
            ]
        for hol_date, hol_suffix in hol_dates:
            self[hol_date] = name + hol_suffix

        # Special holidays (states)
        if year == 2021 and self.subdiv in {"KUL", "LBN", "PJY"}:
            self[date(2021, DEC, 3)] = "Malaysia Cup Holiday"

        if year == 2022 and self.subdiv == "KDH":
            self[date(2022, JAN, 18)] = "Thaipusam"

        if year == 2022 and self.subdiv in {
            "JHR",
            "KDH",
            "KTN",
            "TRG",
        }:
            self[date(2022, MAY, 4)] = "Labour Day Holiday"

        # ---------------------------------#
        # State holidays (multiple states) #
        # ---------------------------------#

        # Good Friday.
        if self.subdiv in {"SBH", "SWK"}:
            self[easter(year) + rd(days=-2)] = "Good Friday"

        # -----------------------------
        # State holidays (single state)
        # -----------------------------

        if self.subdiv == "JHR":
            if year >= 2015:
                self[date(year, MAR, 23)] = "Birthday of the Sultan of Johor"
            if year >= 2011:
                dates_obs = {
                    2011: ((JAN, 12),),
                    2012: ((JAN, 1), (DEC, 20)),
                    2013: ((DEC, 10),),
                    2014: ((NOV, 29),),
                    2015: ((NOV, 19),),
                    2016: ((NOV, 7),),
                    2017: ((OCT, 27),),
                    2018: ((OCT, 15),),
                    2019: ((OCT, 5),),
                    2020: ((SEP, 24),),
                    2021: ((SEP, 13),),
                    2022: ((SEP, 3),),
                }
                name = "Hari Hol of Sultan Iskandar of Johor"
                if year in dates_obs:
                    hol_dates = [
                        (date(year, *date_obs), "")
                        for date_obs in dates_obs[year]
                    ]
                else:
                    hol_dates = [
                        (date_obs, estimated_suffix)
                        for date_obs in _islamic_to_gre(year, 2, 6)
                    ]
                for hol_date, hol_suffix in hol_dates:
                    self[hol_date] = name + hol_suffix

        elif self.subdiv == "KDH":
            if year >= 2020:
                third_sun_jun = date(year, JUN, 1) + rd(weekday=SU(+3))
                self[third_sun_jun] = "Birthday of The Sultan of Kedah"

        elif self.subdiv == "KTN":
            if year >= 2010:
                name = "Birthday of the Sultan of Kelantan"
                self[date(year, NOV, 11)] = name
                self[date(year, NOV, 12)] = name

        elif self.subdiv == "PRK":
            # This Holiday used to be on 27th until 2017
            # https://www.officeholidays.com/holidays/malaysia/birthday-of-the-sultan-of-perak  # noqa: E501
            if year >= 2018:
                first_fri_nov = date(year, NOV, 1) + rd(weekday=FR)
                self[first_fri_nov] = "Birthday of the Sultan of Perak"
            else:
                self[date(year, NOV, 27)] = "Birthday of the Sultan of Perak"

        elif self.subdiv == "SBH":
            first_sat_oct = date(year, OCT, 1) + rd(weekday=SA)
            self[first_sat_oct] = "Birthday of the Governor of Sabah"
            if year >= 2019:
                self[date(year, DEC, 24)] = "Christmas Eve"


class MY(Malaysia):

    # __init__ required for IDE typing and inheritance of docstring.
    def __init__(
        self,
        years: Optional[Union[int, Iterable[int]]] = None,
        expand: bool = True,
        observed: bool = True,
        subdiv: Optional[str] = None,
        prov: Optional[str] = None,
        state: Optional[str] = None,
    ) -> None:
        super().__init__(years, expand, observed, subdiv, prov, state)


class MYS(Malaysia):

    # __init__ required for IDE typing and inheritance of docstring.
    def __init__(
        self,
        years: Optional[Union[int, Iterable[int]]] = None,
        expand: bool = True,
        observed: bool = True,
        subdiv: Optional[str] = None,
        prov: Optional[str] = None,
        state: Optional[str] = None,
    ) -> None:
        super().__init__(years, expand, observed, subdiv, prov, state)



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, JUN, OCT, DEC
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class Nigeria(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Nigeria
    """

    country = "NG"

    def _populate(self, year):
        def _add_holiday(dt: date, hol: str) -> None:
            """Only add if in current year; prevents adding holidays across
            years (handles multi-day Islamic holidays that straddle
            Gregorian years).
            """
            if dt.year == year:
                self[dt] = hol

        if year <= 1978:
            return
        super()._populate(year)

        # New Year's Day
        self[date(year, JAN, 1)] = "New Year's day"

        # Calculate Easter for given year
        # followed by easter related holidays
        easter_date = easter(year)
        self[easter_date + rd(days=-2)] = "Good Friday"
        self[easter_date + rd(days=+1)] = "Easter Monday"

        # Worker's day
        if year >= 1981:
            self[date(year, MAY, 1)] = "Workers' day"

        # Eid al-Fitr - Feast Festive
        # This is an estimate
        # date of observance is announced yearly
        for yr in (year - 1, year):
            for hol_date in _islamic_to_gre(yr, 10, 1):
                _add_holiday(hol_date, "Eid al-Fitr")
                _add_holiday(hol_date + rd(days=1), "Eid al-Fitr Holiday")

        # Arafat Day & Eid al-Adha - Scarfice Festive
        # This is an estimate
        # date of observance is announced yearly
        for yr in (year - 1, year):
            for hol_date in _islamic_to_gre(yr, 12, 10):
                _add_holiday(hol_date, "Eid al-Adha")
                _add_holiday(hol_date + rd(days=1), "Eid al-Adha Holiday")

        # Birthday of Prophet Muhammad
        for hol_date in _islamic_to_gre(year, 3, 12):
            _add_holiday(hol_date, "Mawlid")

        # Independence Day
        self[date(year, OCT, 1)] = "Independence day"

        # Christmas day
        self[date(year, DEC, 25)] = "Christmas day"

        # Boxing day
        self[date(year, DEC, 26)] = "Boxing day"

        # Democracy day moved around after its inception in 2000
        # Initally it fell on May 29th
        # In 2018 it was announced that the holiday
        # will move to June 12th from 2019
        if year >= 2019:
            self[date(year, JUN, 12)] = "Democracy day"
        elif year >= 2000:
            self[date(year, MAY, 29)] = "Democracy day"

        # Observed holidays
        if self.observed and year >= 2016:
            for k, v in list(self.items()):
                if self._is_weekend(k) and k.year == year:
                    next_workday = k + rd(days=+1)
                    while self._is_weekend(next_workday) or self.get(
                        next_workday
                    ):
                        next_workday += rd(days=+1)
                    self[next_workday] = v + " (Observed)"


class NG(Nigeria):
    pass


class NGA(Nigeria):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO, SU
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAY, JUN, AUG, OCT, DEC, SUN
from holidays.holiday_base import HolidayBase


class Jamaica(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Jamaica
    """

    country = "JM"

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        name = "New Year's Day"
        _date = date(year, JAN, 1)
        if self.observed and _date.weekday() == SUN:
            self[_date + rd(weekday=MO(+1))] = name + " (Observed)"
        else:
            self[_date] = name

        # Valentine's Day
        self[date(year, FEB, 14)] = "Valentine's Day"

        # Mother's Day
        self[date(year, MAY, 1) + rd(weekday=SU(+2))] = "Mother's Day"

        # Labour Day
        name = "Labour Day"
        _date = date(year, MAY, 23)
        if self.observed and _date.weekday() == SUN:
            self[_date + rd(weekday=MO)] = name + " (Observed)"
        else:
            self[_date] = name

        # Father's Day
        self[date(year, JUN, 1) + rd(weekday=SU(+3))] = "Father's Day"

        # Emancipation Day
        name = "Emancipation Day"
        _date = date(year, AUG, 1)
        if self.observed and _date.weekday() == SUN:
            self[_date + rd(weekday=MO)] = name + " (Observed)"
        else:
            self[_date] = name

        # Independence Day
        name = "Independence Day"
        _date = date(year, AUG, 6)
        if self.observed and self._is_weekend(_date):
            self[_date + rd(weekday=MO)] = name
        else:
            self[_date] = name

        # National Heroes Day
        self[date(year, OCT, 1) + rd(weekday=MO(+3))] = "National Heroes Day"

        # Christmas
        self[date(year, DEC, 25)] = "Christmas day"

        # Boxing day
        self[date(year, DEC, 26)] = "Boxing day"

        # New Year Eve
        # self[date(year, DEC, 31)] = "New Year Eve"

        # Holidays based on Easter
        easter_date = easter(year)

        # Ash Wednesday
        self[easter_date + rd(days=-46)] = "Ash Wednesday"

        # Good Friday
        self[easter_date + rd(days=-2)] = "Good Friday"

        # Easter
        self[easter_date] = "Easter"

        # Easter
        self[easter_date + rd(days=+1)] = "Easter Monday"


class JM(Jamaica):
    pass


class JAM(Jamaica):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from holidays.constants import JAN, MAR, APR, MAY, JUL, SEP, DEC
from holidays.holiday_base import HolidayBase


class Armenia(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Armenia
    http://www.parliament.am/legislation.php?sel=show&ID=1274&lang=arm&enc=utf8
    https://www.arlis.am/documentview.aspx?docid=259
    """

    country = "AM"

    def _populate(self, year):

        if year <= 1990:
            return
        super()._populate(year)

        # New Year's Day
        name = "Ամանոր [New Year's Day]"
        self[date(year, JAN, 1)] = name
        self[date(year, JAN, 2)] = name

        # Christmas and Epiphany Day
        self[
            date(year, JAN, 6)
        ] = "Սուրբ Ծնունդ եւ Հայտնություն [Christmas and Epiphany Day]"

        if 2010 <= year <= 2021:
            # New Year's holidays and Christmas Eve
            self[date(year, JAN, 3)] = name
            self[date(year, JAN, 4)] = name
            self[date(year, JAN, 5)] = "նախածննդյան տոներ [Christmas Eve]"

            # The Day of Remembrance of the Dead
            self[
                date(year, JAN, 7)
            ] = "Մեռելոց հիշատակի օր [The Day of Remembrance of the Dead]"

        # Army Day
        if year >= 2003:
            self[date(year, JAN, 28)] = "Բանակի օր [Army Day]"

        # Women's Day
        self[date(year, MAR, 8)] = "Կանանց տոն [Women's Day]"

        # Motherhood and Beauty Day
        if 1994 <= year <= 2001:
            self[
                date(year, APR, 7)
            ] = "Մայրության և գեղեցկության տոն [Motherhood and Beauty Day]"

        # Armenian Genocide Remembrance Day
        self[
            date(year, APR, 24)
        ] = "Եղեռնի զոհերի հիշատակի օր [Armenian Genocide Remembrance Day]"

        # Labour Day
        if year >= 2001:
            name = (
                "Աշխատավորների համերաշխության միջազգային օր "
                "[International Day of Workers' Solidarity]"
            )
            if year >= 2002:
                name = "Աշխատանքի օր [Labour Day]"
            self[date(year, MAY, 1)] = name

        # Victory and Peace Day
        if year >= 1995:
            self[
                date(year, MAY, 9)
            ] = "Հաղթանակի և Խաղաղության տոն [Victory and Peace Day]"

        # Republic Day
        self[date(year, MAY, 28)] = "Հանրապետության օր [Republic Day]"

        # Constitution Day
        if year >= 1996:
            self[date(year, JUL, 5)] = "Սահմանադրության օր [Constitution Day]"

        # Independence Day
        if year >= 1992:
            self[date(year, SEP, 21)] = "Անկախության օր [Independence Day]"

        # New Year's Eve
        self[date(year, DEC, 31)] = "Ամանոր [New Year's Eve]"


class AM(Armenia):
    pass


class ARM(Armenia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, MAY, JUN, JUL, AUG, OCT, NOV, DEC, MON
from holidays.holiday_base import HolidayBase


class Colombia(HolidayBase):
    """
    Colombia has 18 holidays. The establishing of these are by:
    Ley 35 de 1939 (DEC 4): https://bit.ly/3PJwk7B
    Decreto 2663 de 1950 (AUG 5): https://bit.ly/3PJcut8
    Decreto 3743 de 1950 (DEC 20): https://bit.ly/3B9Otr3
    Ley 51 de 1983 (DEC 6): https://bit.ly/3aSobiB
    """

    country = "CO"

    def _add_with_bridge(self, _date, name):
        """
        On the 6th of December 1983, the government of Colombia declared which
        holidays are to take effect, and also clarified that a subset of them
        are to take place the next Monday if they do not fall on a Monday.
        This law is "Ley 51 de 1983" which translates to law 51 of 1983.
        Link: https://bit.ly/3PtPi2e
        A few links below to calendars from the 1980s to demonstrate this law
        change. In 1984 some calendars still use the old rules, presumably
        because they were printed prior to the declaration of law change.
        1981: https://bit.ly/3BbgKOc
        1982: https://bit.ly/3BdbhWW
        1984: https://bit.ly/3PqGxWU
        1984: https://bit.ly/3B7ogt8
        """

        if self.observed and _date.weekday() != MON and _date.year > 1983:
            self[_date + rd(weekday=MO)] = name + " (Observed)"
        else:
            self[_date] = name

    def _populate(self, year):
        super()._populate(year)

        self._add_fixed_date_holidays(year)
        self._add_flexible_date_holidays(year)
        self._add_easter_based_holidays(year)

    def _add_fixed_date_holidays(self, year):
        """
        These holidays are always on the same date no matter what day of the
        week they fall on.
        """

        # New Year's Day
        self[date(year, JAN, 1)] = "Año Nuevo [New Year's Day]"

        # Labor Day
        self[date(year, MAY, 1)] = "Día del Trabajo [Labour Day]"

        # Independence Day
        self[
            date(year, JUL, 20)
        ] = "Día de la Independencia [Independence Day]"

        # Battle of Boyaca
        self[date(year, AUG, 7)] = "Batalla de Boyacá [Battle of Boyacá]"

        if year > 1950:
            # Immaculate Conception
            self[
                date(year, DEC, 8)
            ] = "La Inmaculada Concepción [Immaculate Conception]"

        # Christmas
        self[date(year, DEC, 25)] = "Navidad [Christmas]"

    def _add_flexible_date_holidays(self, year):
        """
        These holidays fall on the next Monday if they are not already on
        Monday.
        """

        if year > 1950:
            # Epiphany
            self._add_with_bridge(
                date(year, JAN, 6),
                "Día de los Reyes Magos [Epiphany]",
            )

            # Saint Joseph's Day
            self._add_with_bridge(
                date(year, MAR, 19),
                "Día de San José [Saint Joseph's Day]",
            )

            # Saint Peter and Saint Paul's Day
            self._add_with_bridge(
                date(year, JUN, 29),
                "San Pedro y San Pablo [Saint Peter and Saint Paul]",
            )

            # Assumption of Mary
            self._add_with_bridge(
                date(year, AUG, 15),
                "La Asunción [Assumption of Mary]",
            )

        # Columbus Day
        self._add_with_bridge(
            date(year, OCT, 12),
            "Día de la Raza [Columbus Day]",
        )

        if year > 1950:
            # All Saints’ Day
            self._add_with_bridge(
                date(year, NOV, 1),
                "Día de Todos los Santos [All Saint's Day]",
            )

        # Independence of Cartagena
        self._add_with_bridge(
            date(year, NOV, 11),
            "Independencia de Cartagena [Independence of Cartagena]",
        )

    def _add_easter_based_holidays(self, year):
        """
        These holidays change each year based on when easter is.
        """

        _easter = easter(year)
        self._add_fixed_easter_based_holidays(_easter)
        self._add_flexible_easter_based_holidays(_easter)

    def _add_fixed_easter_based_holidays(self, _easter):
        if _easter.year > 1950:
            # Maundy Thursday
            self[_easter + rd(days=-3)] = "Jueves Santo [Maundy Thursday]"

            # Good Friday
            self[_easter + rd(days=-2)] = "Viernes Santo [Good Friday]"

    def _add_flexible_easter_based_holidays(self, _easter):
        if _easter.year > 1950:
            # Ascension of Jesus
            self._add_with_bridge(
                _easter + rd(days=+39),
                "Ascensión del señor [Ascension of Jesus]",
            )

            # Corpus Christi
            self._add_with_bridge(
                _easter + rd(days=+60),
                "Corpus Christi [Corpus Christi]",
            )

        if _easter.year > 1983:
            # Sacred Heart
            self._add_with_bridge(
                _easter + rd(days=+68),
                "Sagrado Corazón [Sacred Heart]",
            )


class CO(Colombia):
    pass


class COL(Colombia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, APR, MAY, JUN, JUL, AUG, OCT, NOV, DEC
from holidays.constants import TUE, WED, THU, FRI
from holidays.holiday_base import HolidayBase


class Uruguay(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Uruguay
    """

    country = "UY"

    def _populate(self, year):
        super()._populate(year)

        # Mandatory paid holidays:

        # New Year's Day.
        self[date(year, JAN, 1)] = "Año Nuevo [New Year's Day]"

        # Día de los Trabajadores.
        self[
            date(year, MAY, 1)
        ] = "Día de los Trabajadores [International Workers' Day]"

        # Jura de la Constitución.
        self[
            date(year, JUL, 18)
        ] = "Jura de la Constitución [Constitution Day]"

        # Declaratoria de la Independencia.
        self[
            date(year, AUG, 25)
        ] = "Día de la Independencia [Independence Day]"

        # Christmas.
        self[date(year, DEC, 25)] = "Día de la Familia [Day of the Family]"

        # Partially paid holidays:

        # Día de los Niños.
        self[date(year, JAN, 6)] = "Día de los Niños [Children's Day]"

        # Natalicio de José Gervasio Artigas.
        self[date(year, JUN, 19)] = (
            "Natalicio de José Gervasio Artigas "
            "[Birthday of José Gervasio Artigas]"
        )

        # Día de los difuntos.
        self[date(year, NOV, 2)] = "Día de los Difuntos [All Souls' Day]"

        # Moveable holidays:

        easter_date = easter(year)

        # Carnival days
        # revisar este día para futuros casos
        name = "Día de Carnaval [Carnival's Day]"
        self[easter_date + rd(days=-48)] = name
        self[easter_date + rd(days=-47)] = name

        # Holy Week.
        self[easter_date + rd(days=-3)] = "Jueves Santo [Holy Thursday]"
        self[easter_date + rd(days=-2)] = "Viernes Santo [Holy Friday]"

        self[easter_date] = "Día de Pascuas [Easter Day]"

        holiday_pairs = (
            (
                # Desembarco de los 33 Orientales en la playa de la Agraciada.
                date(year, APR, 19),
                "Desembarco de los 33 Orientales [Landing of the 33 Patriots]",
            ),
            (
                # Batalla de Las Piedras [Battle of Las Piedras].
                date(year, MAY, 18),
                "Batalla de Las Piedras [Battle of Las Piedras]",
            ),
            (
                # "Día del Respeto a la Diversidad Cultural
                # [Respect for Cultural Diversity Day].
                date(year, OCT, 12),
                "Día del Respeto a la Diversidad Cultural "
                "[Respect for Cultural Diversity Day]",
            ),
        )

        for dt, name in holiday_pairs:
            if dt.weekday() in {TUE, WED}:
                self[dt + rd(weekday=MO(-1))] = name
            elif dt.weekday() in {THU, FRI}:
                self[dt + rd(weekday=MO(+1))] = name
            else:
                self[dt] = name


class UY(Uruguay):
    pass


class URY(Uruguay):
    pass
